# -*- coding: utf-8 -*-
import sys

from lumbermill.constants import TYPENAMES_TO_TYPE
from lumbermill.utils.DynamicValues import replaceVarsAndCompileString

if sys.hexversion > 0x03000000:
    pass
else:
    pass

yaml_valid_config_template = {
    'Global': {'types': [dict],
               'fields': {'workers': {'types': [int]}}},
    'Module': {'types': [dict,str],
               'fields': {'id': {'types': [str]},
                          'filter': {'types': [str]},
                          'add_fields': {'types': [dict]},
                          'delete_fields': {'types': [list]},
                          'event_type': {'types': [str]},
                          'receivers': {'types': [list]}}
               }
}

class ConfigurationValidator():
    """
    Validate a module instance based on the configuration meta data retrieved from the docstring via the ModuleDocstringParser.

    The docstring should contain patterns like:
    ...
    compression: 'zlib' # <default: 'gzip'; type: string; is: optional; values: [gzip, zlib]>
    ...

    Where:
    - target_field ist the configuration key
    - default: sets a standard value
    - type: sets the variable type
    - is: sets whether the parameter needs to be provided or not.
      Here a simple conditional syntax is supported, e.g.
      is: required if tls is True else optional
    - values (optional): sets a list of allowed values.
    """

    default_module_config_keys = ('module', 'id', 'filter', 'receivers', 'pool_size', 'queue_size', 'mp_queue_buffer_size','redis_store', 'redis_key', 'redis_ttl', 'add_fields', 'delete_fields', 'event_type')

    @classmethod
    def validateConfiguration(self, configuration_data):
        """
        Simple schema test for the global configuration.

        Only a very simple "schema" is checked here.
        Global configuration item and each module configuration item should at least adhere to the data pattern defined in
        yaml_valid_config_template. Module specific configuration checks will be done in validateModuleInstanceConfiguration.
        """
        configuration_errors = []
        for configuration_item in configuration_data:
            if type(configuration_item) is str:
                # Simple modules names are ok.
                continue
            elif type(configuration_item) is list:
                # List items in root configuraion are not allowed.
                error_msg = "'%s'(list) is not allowed here. Please check your configuration." % configuration_item
                configuration_errors.append(error_msg)
                continue
            elif type(configuration_item) is dict:
                for key, value in configuration_item.items():
                    mapped_key = 'Module' if key != 'Global' else key
                    item_configuration_errors = self.validateConfigurationItem(mapped_key, value, key)
                    for item_configuration_error in item_configuration_errors:
                        configuration_errors.append(item_configuration_error)
                continue
            else:
                error_msg = "'%s' is of invalid type %s. Please check your configuration." % (configuration_item, type(configuration_item))
                configuration_errors.append(error_msg)
        return configuration_errors

    @classmethod
    def validateConfigurationItem(self, item_name, item_value, path, template=yaml_valid_config_template):
        configuration_errors = []
        if item_name in template:
            item_template = template[item_name]
            if type(item_value) not in item_template['types']:
                error_msg = "'%s' not of correct datatype. Is: %s, should be: %s. Please check your configuration." % (path, type(item_value), item_template['types'])
                configuration_errors.append(error_msg)
            if type(item_value) is dict:
                for field_key, field_value in item_value.items():
                    field_path = "%s.%s" % (path, field_key)
                    try:
                        field_configuration_errors = self.validateConfigurationItem(field_key, field_value, field_path, template=item_template['types'])
                    except KeyError:
                        sys.exit()
                    for field_configuration_error in field_configuration_errors:
                        configuration_errors.append(field_configuration_error)
        return configuration_errors

    def validateModuleConfiguration(self, moduleInstance):
        result = []
        # The ModifyFields module is an exception as it provides more than one configuration.
        # This needs to be taken into account when testing for required configuration values.
        # At the moment, I just skip this module until I have a good idea on how to tackle this.
        if not hasattr(moduleInstance, 'configuration_metadata') or moduleInstance.__class__.__name__ in ['ModifyFields']:
            return result
        # Check for pool_size > 1 in single threaded/processed modules.
        if 'pool_size' in moduleInstance.configuration_data and moduleInstance.getConfigurationValue('pool_size') > 1 and moduleInstance.can_run_forked is False:
            error_msg = "%s: 'pool_size' has invalid value. Is: %s but the module can only run in one thread/process." % (moduleInstance.__class__.__name__, moduleInstance.getConfigurationValue('pool_size'))
            result.append(error_msg)
        # Check if the live configuration provides a key that is not documented in modules docstring.
        config_keys_not_in_docstring = set(moduleInstance.configuration_data.keys()) - set(moduleInstance.configuration_metadata.keys()) - set(self.default_module_config_keys)
        if config_keys_not_in_docstring:
            keys = []
            for key in config_keys_not_in_docstring:
                keys.append(key)
            error_msg = "%s: Found unknown configuration keys: %s. Please check module documentation." % (moduleInstance.__class__.__name__, keys)
            result.append(error_msg)
            return result
        for configuration_key, configuration_metadata in moduleInstance.configuration_metadata.items():
            config_value = moduleInstance.getConfigurationValue(configuration_key)
            config_value_datatype = type(config_value)
            # Check for required parameter.
            if 'is' in configuration_metadata:
                dependency = configuration_metadata['is']
                edits = [('optional', '"optional"'), ('required', '"required"')]
                for search, replace in edits:
                    dependency = dependency.replace(search, replace)
                try:
                    # TODO: Think of a better and more secure way to evaluate the dependencies.
                    exec(replaceVarsAndCompileString("dependency = %s" % dependency, 'moduleInstance.getConfigurationValue("%s")'))
                except TypeError:
                    etype, evalue, etb = sys.exc_info()
                    error_msg = "%s: Could not parse dependency %s in '%s'. Exception: %s, Error: %s." % (dependency, moduleInstance.__class__.__name__, etype, evalue, configuration_key)
                    result.append(error_msg)
                if dependency == 'required' and not config_value:
                    error_msg = "%s: '%s' not configured but is required. Please check module documentation." % (moduleInstance.__class__.__name__, configuration_key)
                    result.append(error_msg)
                    continue
            # Check for value type.
            allowed_datatypes = []
            if 'type' in configuration_metadata:
                for allowed_datatypes_as_string in configuration_metadata['type']:
                    try:
                        allowed_datatypes.append(TYPENAMES_TO_TYPE[allowed_datatypes_as_string.title()])
                    except KeyError:
                        error_msg = "%s: Docstring config setting for '%s' has unknown datatype '%s'. Supported datatypes: %s" % (moduleInstance.__class__.__name__, configuration_key, allowed_datatypes_as_string.title(), TYPENAMES_TO_TYPE.keys())
                        result.append(error_msg)
                if config_value_datatype not in allowed_datatypes:
                    error_msg = "%s: '%s' not of correct datatype. Is: %s, should be: %s" % (moduleInstance.__class__.__name__, configuration_key, config_value_datatype, allowed_datatypes)
                    result.append(error_msg)
            # Check for value restrictions.
            if 'values' in configuration_metadata:
                if config_value not in configuration_metadata['values']:
                    error_msg = "%s: '%s' has invalid value. Is: %s, should be one of: %s" % (moduleInstance.__class__.__name__, configuration_key, config_value, configuration_metadata['values'])
                    result.append(error_msg)
        return result


# -*- coding: utf-8 -*-
import logging
import socket
import sys
import time

import pylru

try:
    import msgpack
    msgpack_avaiable = True
except ImportError:
    msgpack_avaiable = False

try:
    import zmq
    zmq_avaiable = True
except ImportError:
    zmq_avaiable = False

from lumbermill.utils.Decorators import setInterval
from lumbermill.utils.misc import TimedFunctionManager
from lumbermill.utils.DictUtils import KeyDotNotationDict

class Buffer:
    def __init__(self, flush_size=None, callback=None, interval=1, maxsize=5000):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.flush_size = flush_size
        self.buffer = []
        self.maxsize = maxsize
        self.append = self.put
        self.flush_interval = interval
        self.flush_callback = callback
        self.flush_timed_func = self.getTimedFlushMethod()
        self.timed_func_handle = TimedFunctionManager.startTimedFunction(self.flush_timed_func)
        self.is_flushing = False

    def stopInterval(self):
        TimedFunctionManager.stopTimedFunctions(self.timed_func_handle)
        self.timed_func_handle = False

    def startInterval(self):
        if self.timed_func_handle:
            self.stopInterval()
        self.timed_func_handle = TimedFunctionManager.startTimedFunction(self.flush_timed_func)

    def getTimedFlushMethod(self):
        @setInterval(self.flush_interval)
        def timedFlush():
            self.flush()
        return timedFlush

    def append(self, item):
        self.put(item)

    def put(self, item):
        # Wait till a running store is finished to avoid strange race conditions when using this buffer with multiprocessing.
        while self.is_flushing:
            time.sleep(.00001)
        while len(self.buffer) > self.maxsize:
            self.logger.warning("Maximum number of items (%s) in buffer reached. Waiting for flush." % self.maxsize)
            time.sleep(1)
        self.buffer.append(item)
        if self.flush_size and len(self.buffer) == self.flush_size:
            self.flush()

    def flush(self):
        if self.bufsize() == 0 or self.is_flushing:
            return
        self.is_flushing = True
        self.stopInterval()
        success = self.flush_callback(self.buffer)
        if success:
            self.buffer = []
        self.startInterval()
        self.is_flushing = False

    def bufsize(self):
        return len(self.buffer)

class BufferedQueue:
    def __init__(self, queue, buffersize=500):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.queue = queue
        self.buffersize = buffersize
        self.buffer = Buffer(buffersize, self.sendBuffer, 5)

    def startInterval(self):
        self.buffer.startInterval()

    def put(self, payload):
        self.buffer.append(payload)

    def sendBuffer(self, buffered_data):
        try:
            buffered_data = msgpack.packb(buffered_data)
            self.queue.put(buffered_data)
            return True
        except (KeyboardInterrupt, SystemExit):
            # Keyboard interrupt is catched in GambolPuttys main run method.
            # This will take care to shutdown all running modules.
            pass
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not append data to queue. Exception: %s, Error: %s." % (etype, evalue))

    def get(self, block=True, timeout=None):
        try:
            buffered_data = self.queue.get(block, timeout)
            buffered_data = msgpack.unpackb(buffered_data)
            # After msgpack.uppackb we just have a normal dict. Cast this to KeyDotNotationDict.
            for data in buffered_data:
                yield KeyDotNotationDict(data)
        except (KeyboardInterrupt, SystemExit, ValueError, OSError):
            # Keyboard interrupt is catched in GambolPuttys main run method.
            # This will take care to shutdown all running modules.
            pass

    def qsize(self):
        return self.buffer.bufsize() + self.queue.qsize()

    def __getattr__(self, name):
        return getattr(self.queue, name)

class ZeroMqMpQueue:
    """
    Use ZeroMQ for IPC.
    This is faster than the default multiprocessing.Queue.

    Sender and receiver will be initalized on first put/get. This is neccessary since a zmq context will not
    survive a fork.

    send_pyobj and recv_pyobj is not used since it performance is slower than using msgpack for serialization.
    (A test for a simple dict using send_pyobj et.al performed around 12000 eps, while msgpack and casting to
    KeyDotNotationDict after unpacking resulted in around 17000 eps)
    """
    def __init__(self, queue_max_size=20):
        # Get a free random port.
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(('127.0.0.1', 0))
        sock.listen(socket.SOMAXCONN)
        ipaddr, self.port = sock.getsockname()
        sock.close()
        self.queue_max_size = queue_max_size
        self.queue_size = 0
        self.sender = None
        self.receiver = None

    def initSender(self):
        zmq_context = zmq.Context()
        self.sender = zmq_context.socket(zmq.PUSH)
        try:
            self.sender.setsockopt(zmq.SNDHWM, self.queue_max_size)
        except AttributeError:
            self.sender.setsockopt(zmq.HWM, self.queue_max_size)
        try:
            self.sender.bind("tcp://127.0.0.1:%d" % self.port)
        except zmq.error.ZMQError:
            print("Address in use. Connecting only.")
            self.sender.connect("tcp://127.0.0.1:%d" % self.port)


    def initReceiver(self):
        #print("Init receiver in %s" % os.getpid())
        zmq_context = zmq.Context()
        self.receiver = zmq_context.socket(zmq.PULL)
        try:
            self.receiver.setsockopt(zmq.RCVHWM, self.queue_max_size)
        except AttributeError:
            self.receiver.setsockopt(zmq.HWM, self.queue_max_size)
        self.receiver.connect("tcp://127.0.0.1:%d" % self.port)

    def put(self, data):
        if not self.sender:
            self.initSender()
        self.sender.send(data)

    def get(self, block=None, timeout=None):
        if not self.receiver:
            self.initReceiver()
        events = ""
        try:
            events = self.receiver.recv()
            return events
        except zmq.error.ZMQError as e:
            # Ignore iterrupt error caused by SIGINT
            if e.strerror == "Interrupted system call":
                return events

    def qsize(self):
        return self.queue_size

class MemoryCache():

    def __init__(self, size=1000):
        self.lru_dict = pylru.lrucache(size)

    def set(self, key, value):
        self.lru_dict[key] = value

    def get(self, key):
        return self.lru_dict[key]

    def unset(self, key):
        return self.lru_dict.pop(key)


# -*- coding: utf-8 -*-
from tornado.web import UIModule
from pygments import highlight
from pygments.lexers import JsonLexer
from pygments.formatters import HtmlFormatter
import collections


class ServerConfiguration(UIModule):
    def render(self, webserver_module, render_type="text"):
        print self.__dict__
        if render_type == "text":
            return self.render_string("server_configuration_text.html", node=node)
        else:
            return self.render_string("server_configuration_graph.html", node=node)

    def getServerConfiguration(self, webserver_module):
        modules_info = collections.OrderedDict()
        for module_id, module_info in sorted(webserver_module.gp.modules.items(), key=lambda x: x[1]['idx']):
            modules_info[module_id] = {'id': module_id, 'type': module_info['type'], 'configuration': module_info['configuration']}


# -*- coding: utf-8 -*-
import sys
import types

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class CollectdParser(BaseThreadedModule):
    """
    Parse collectd binary protocol data.

    This module can receive binary data from the collectd network plugin.

    Decode:
      It will parse the collectd binary data and create or replace fields in the internal data dictionary with
      the corresponding collectd data.
    Encode:
      Encode selected fields or all to collectd binary protocol.

    Configuration template:

    - CollectdParser:
       action:                          # <default: 'decode'; type: string; values: ['decode','encode']; is: optional>
       source_fields:                   # <default: 'data'; type: string||list; is: optional>
       target_field:                    # <default: None; type: None||string; is: optional>
       keep_original:                   # <default: False; type: boolean; is: optional>
       receivers:
        - NextModule
    """

    module_type = "parser"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.source_fields = self.getConfigurationValue('source_fields')
        # Allow single string as well.
        if isinstance(self.source_fields, types.StringTypes):
            self.source_fields = [self.source_fields]
        self.target_field = self.getConfigurationValue('target_field')
        self.drop_original = not self.getConfigurationValue('keep_original')
        if self.getConfigurationValue('action') == 'decode':
            self.default_list_attributes = [attrib for attrib in dir(list) if not attrib.startswith('__')]
            self.parser = Parser()
            self.handleEvent = self.decodeEvent
        else:
            self.handleEvent = self.encodeEvent

    def decodeEvent(self, event):
        for source_field in self.source_fields:
            if source_field not in event:
                continue
            try:
                decoded_data = self.parser.interpret(event[source_field])
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.warning("Could not decode event data: %s. Exception: %s, Error: %s." % (event[source_field], etype, evalue))
                continue
            collectd_values = {}
            for value in decoded_data:
                all_attibutes = [attrib for attrib in dir(value) if not attrib.startswith('__')]
                for attribute in all_attibutes:
                    if attribute not in self.default_list_attributes:
                        collectd_values[attribute] = getattr(value, attribute)
            if self.drop_original:
                event.pop(source_field, None)
            if self.target_field:
                event.update({self.target_field: collectd_values})
            else:
                try:
                    event.update(collectd_values)
                except:
                    etype, evalue, etb = sys.exc_info()
                    self.logger.warning("Could not update event with collectd data: %s. Exception: %s, Error: %s." % (decoded_data, etype, evalue))
        yield event

    def encodeEvent(self, event):
        # TODO: implement encoding
        pass

# Copyright © 2009 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv2 license.

#
# Updated by Rami Sayar for Collectd 5.1. Added DERIVE handling.
# Updated by Grégory Starck with few enhancements.
#  - notably possibility to subclass Values and Notification.

"""
Collectd network protocol implementation.
"""

import struct

from datetime import datetime
from copy import deepcopy


#############################################################################s

# https://collectd.org/wiki/index.php/Binary_protocol
# -> Protocol structure
# " The maximum length of payload in any part is therefore 65531 bytes. "
_BUFFER_SIZE = 65535 # 65535 > 65531, ok we are safe.

#############################################################################s

class CollectdException(Exception):
    pass

class CollectdValueError(CollectdException, ValueError):
    pass

class CollectdDecodeError(CollectdValueError):
    pass

class CollectdUnsupportedDSType(CollectdValueError):
    pass

class CollectdUnsupportedMessageType(CollectdValueError):
    pass

class CollectdBufferOverflow(CollectdValueError):
    pass

#############################################################################s

# Message kinds
TYPE_HOST            = 0x0000
TYPE_TIME            = 0x0001
TYPE_PLUGIN          = 0x0002
TYPE_PLUGIN_INSTANCE = 0x0003
TYPE_TYPE            = 0x0004
TYPE_TYPE_INSTANCE   = 0x0005
TYPE_VALUES          = 0x0006
TYPE_INTERVAL        = 0x0007
TYPE_TIMEHR          = 0x0008
TYPE_INTERVALHR      = 0x0009

# For notifications
TYPE_MESSAGE         = 0x0100
TYPE_SEVERITY        = 0x0101

#
TYPE_SIGN_SHA256     = 0x0200
TYPE_ENCR_AES256     = 0x0210


# DS kinds
DS_TYPE_COUNTER      = 0
DS_TYPE_GAUGE        = 1
DS_TYPE_DERIVE       = 2
DS_TYPE_ABSOLUTE     = 3

header = struct.Struct("!2H")
number = struct.Struct("!Q")
signed_number = struct.Struct("!q") # DERIVE are signed long longs
short  = struct.Struct("!H")
double = struct.Struct("<d")


assert double.size == number.size == signed_number.size == 8

#############################################################################s

_values_header_size = header.size + short.size
_single_value_size = 1 + 8 # 1 byte for type, 8 bytes for value


_ds_type_decoder = {
    DS_TYPE_COUNTER:    number,
    DS_TYPE_ABSOLUTE:   number,
    DS_TYPE_DERIVE:     signed_number,
    DS_TYPE_GAUGE:      double
}

def decode_network_values(ptype, plen, buf):
    """Decodes a list of DS values in collectd network format
    """
    assert ptype == TYPE_VALUES

    nvalues = short.unpack_from(buf, header.size)[0]
    values_tot_size = _values_header_size + nvalues * _single_value_size
    if values_tot_size != plen:
        raise CollectdValueError('Values total size != Part len (%s vs %s)' % (values_tot_size, plen))

    results = []
    off = _values_header_size + nvalues

    for dstype in map(ord, buf[_values_header_size:off]):
        try:
            decoder = _ds_type_decoder[dstype]
        except KeyError:
            raise CollectdUnsupportedDSType("DS type %i unsupported" % dstype)
        results.append((dstype, decoder.unpack_from(buf, off)[0]))
        off += 8
    return results


def decode_network_number(ptype, plen, buf):
    """Decodes a number (64-bit unsigned) in collectd network format.
    """
    return number.unpack_from(buf, header.size)[0]


def decode_network_string(ptype, plen, buf):
    """Decodes a string (\0 terminated) in collectd network format.
    :return: The string as bytes (not unicode).
    """
    return buf[header.size:plen-1]

# Mapping of message types to decoding functions.
_decoders = {
    TYPE_VALUES         : decode_network_values,
    TYPE_TIME           : decode_network_number,
    TYPE_INTERVAL       : decode_network_number,
    TYPE_HOST           : decode_network_string,
    TYPE_PLUGIN         : decode_network_string,
    TYPE_PLUGIN_INSTANCE: decode_network_string,
    TYPE_TYPE           : decode_network_string,
    TYPE_TYPE_INSTANCE  : decode_network_string,
    TYPE_MESSAGE        : decode_network_string,
    TYPE_SEVERITY       : decode_network_number,
    TYPE_TIMEHR         : decode_network_number,
    TYPE_INTERVALHR     : decode_network_number,
}


def decode_network_packet(buf):
    """Decodes a network packet in collectd format.
    """
    off = 0
    blen = len(buf)

    while off < blen:
        try:
            ptype, plen = header.unpack_from(buf, off)
        except struct.error as err:
            raise CollectdDecodeError(err)
        if not plen:
            raise CollectdValueError('Invalid part with size=0: buflen=%s off=%s ptype=%s' % (
                                      blen, off, ptype))

        rest = blen - off
        if plen > rest:
            raise CollectdBufferOverflow("Encoded part size greater than left amount of data in buffer: buflen=%s off=%s vsize=%s" % (
                blen, off, plen))

        try:
            decoder = _decoders[ptype]
        except KeyError:
            raise CollectdUnsupportedMessageType("Part type %s not recognized (off=%s)" % (ptype, off))

        try:
            res = decoder(ptype, plen, buf[off:])
        except struct.error as err:
            raise CollectdDecodeError(err)

        yield ptype, res
        off += plen

#############################################################################s

def cdtime_to_time(cdt):
    '''
    :param cdt: A CollectD Time or Interval HighResolution encoded value.
    :return: A float, representing a time EPOCH value, with up to nanosec after comma.
    '''
    # fairly copied from http://git.verplant.org/?p=collectd.git;a=blob;f=src/daemon/utils_time.h
    sec = cdt >> 30
    nsec = ((cdt & 0b111111111111111111111111111111) / 1.073741824) / 10**9
    assert 0 <= nsec < 1
    return sec + nsec


class Data(object):
    time = None
    interval = None
    host = None
    plugin = None
    plugininstance = None
    type = None
    typeinstance = None

    def __init__(self, **kw):
        for k,v in kw.iteritems():
            setattr(self, k, v)

    @property
    def datetime(self):
        return datetime.fromtimestamp(self.time)

    @property
    def source(self):
        res = []
        if self.host:
            res.append(self.host)
        for attr in ('plugin', 'plugininstance', 'type', 'typeinstance'):
            val = getattr(self, attr)
            if val:
                res.append("/")
                res.append(val)
        return ''.join(res)

    def __str__(self):
        return "[%s] %s" % (self.time, self.source)



class Notification(Data):
    ''' Notification
    '''
    FAILURE  = 1
    WARNING  = 2
    OKAY     = 4

    SEVERITY = {
        FAILURE: "FAILURE",
        WARNING: "WARNING",
        OKAY   : "OKAY",
    }

    __severity = 0
    message  = ""

    def __set_severity(self, value):
        if value in (self.FAILURE, self.WARNING, self.OKAY):
            self.__severity = value

    severity = property(lambda self: self.__severity, __set_severity)

    @property
    def severitystring(self):
        return self.SEVERITY.get(self.severity, "UNKNOWN")

    def __str__(self):
        return "%s [%s] %s" % (
                super(Notification, self).__str__(),
                self.severitystring,
                self.message)



class Values(Data, list):
    ''' collectd Values ; contains a list of values associated with a particular collectd "element"
    '''
    def __str__(self):
        return "%s %s" % (Data.__str__(self), list.__str__(self))


#############################################################################s

class Parser(object):
    ''' Represent a collectd parser.

    Feed its `interpret´ method with some input and get Values or Notification instances.
    '''
    Values = Values # so to be able to customize their behavior
    Notification = Notification


    def receive(self):
        ''' Method used by the parser to get some data if you don't feed it explicitly with.
        If you want to make use of it you have to subclass and define it respecting the return format.
        :return: a 2-tuple : (buffer_read, address_read_from)
        The address_read_from format isn't enforced.
        '''
        raise NotImplementedError

    def decode(self, buf=None):
        """Decodes a given buffer or the next received packet from `receive()´.
        :return: a generator yielding 2-tuples (type, value).
        """
        if buf is None:
            buf, addr_from = self.receive()
        return decode_network_packet(buf)


    def interpret_opcodes(self, iterable):
        '''
        :param iterable: An iterable of 2-tuples (type, value).
        :return: A generator yielding Values or Notification instances based on the iterable.
        :raise: The generator, when yielding results, can raise a CollectdException (or subclass)
                instance if there is a decode error.
        '''
        vl = self.Values() # ; assert isinstance(vl, Values)
        nt = self.Notification() # ; assert isinstance(nt, Notification)

        for kind, data in iterable:

            if kind == TYPE_TIME:
                vl.time = nt.time = data
            elif kind == TYPE_TIMEHR:
                vl.time = nt.time = cdtime_to_time(data)
            elif kind == TYPE_INTERVAL:
                vl.interval = data
            elif kind == TYPE_INTERVALHR:
                vl.interval = cdtime_to_time(data)
            elif kind == TYPE_HOST:
                vl.host = nt.host = data
            elif kind == TYPE_PLUGIN:
                vl.plugin = nt.plugin = data
            elif kind == TYPE_PLUGIN_INSTANCE:
                vl.plugininstance = nt.plugininstance = data
            elif kind == TYPE_TYPE:
                vl.type = nt.type = data
            elif kind == TYPE_TYPE_INSTANCE:
                vl.typeinstance = nt.typeinstance = data
            elif kind == TYPE_SEVERITY:
                nt.severity = data
            elif kind == TYPE_MESSAGE:
                nt.message = data
                yield deepcopy(nt)
            elif kind == TYPE_VALUES:
                vl[:] = data
                yield deepcopy(vl)
            # others kinds are just ignored for now.
            # it's permitted by collectd packet format so to be extensible.


    def interpret(self, input=None):
        """Interprets an explicit or implicit `input´ "sequence" if given.

        :param input:
            If None or not given -> A fresh packet will be read from the socket. Then the packet will be decode().
            If a basestring -> It will also be decode().

            After what the result of decode() (a generator which yields 2-tuple(value_type, value)) is given
            to interpret_opcodes() which will then yield collectd `Values´ or `Notification´ instances.

            If the `input´ initial value isn't None nor a basestring then it's directly given to
            interpret_opcodes(), you have to make sure the `input´ has the correct format.

        :return: A generator yielding collectd `Values´ or `Notification´ instances.

        :raise:
            When a read on the socket is needed, it's not impossible to raise some IO exception.
            Otherwise no raise should occur to return the generator.
            But the returned generator can raise (subclass-)`CollectdException´ instance if a decode problem occurs.
        """
        if isinstance(input, (type(None), basestring)):
            input = self.decode(input)
        return self.interpret_opcodes(input)


# -*- coding: utf-8 -*-
import sys

from lxml import etree

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class XPathParser(BaseThreadedModule):
    """
    Parse an xml string via xpath.

    This module supports the storage of the results in an redis db. If redis-client is set,
    it will first try to retrieve the result from redis via the key setting.
    If that fails, it will execute the xpath query and store the result in redis.

    Configuration template:

    - XPathParser:
       source_field:                    # <type: string; is: required>
       target_field:                    # <default: "gambolputty_xpath"; type: string; is: optional>
       query:                           # <type: string; is: required>
       redis_store:                     # <default: None; type: None||string; is: optional>
       redis_key:                       # <default: None; type: None||string; is: optional if redis_store is None else required>
       redis_ttl:                       # <default: 60; type: integer; is: optional>
       receivers:
        - NextModule
    """

    module_type = "parser"
    """Set module type"""

    def configure(self, configuration):
        BaseThreadedModule.configure(self, configuration)
        self.redis_ttl = self.getConfigurationValue('redis_ttl')
        # Get redis client module.
        if self.getConfigurationValue('redis_store'):
            mod_info = self.lumbermill.getModuleInfoById(self.getConfigurationValue('redis_store'))
            self.redis_store = mod_info['instances'][0]
        else:
            self.redis_store = None

    def castToList(self, value):
        list = []
        for x in value:
            try:
                list.append(etree.tostring(x))
            except TypeError:
                list.append(str(x))
        return list

    def handleEvent(self, event):
        """
        Process the event.

        @param event: dictionary
        @return data: dictionary
        """
        source_field = self.getConfigurationValue('source_field', event)
        if source_field not in event:
            yield event
            return
        result = None
        if self.redis_store:
            redis_key = self.getConfigurationValue('redis_key', event)
            result = self.redis_store.get(redis_key)
        if result is None:
            xml_string = event[source_field].decode('utf8').encode('ascii', 'ignore')
            try:
                xml_root = etree.fromstring(xml_string)
                xml_tree = etree.ElementTree(xml_root)
                result =  xml_tree.xpath(self.getConfigurationValue('query', event))
                if(type(result) == list):
                    result = self.castToList(result)
                if self.redis_store:
                    self.redis_store.set(redis_key, result, self.redis_ttl)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.warning("Could not parse xml doc %s Exception: %s, Error: %s." % (xml_string, etype, evalue))
        if result:
            target_field_name = self.getConfigurationValue('target_field', event)
            event[target_field_name] = result
        yield event


# -*- coding: utf-8 -*-
import sys
import os
import signal
import subprocess


class TimedFunctionManager:
    """
    The decorator setInterval provides a simple way to repeatedly execute a function in intervals.
    This is done by starting a thread that calls the decorated method every interval seconds.
    To make sure, all threads get stopped when exiting or reloading LumberMill, the decorated functions
    should be started like this e.g.:
    ...
    Utils.TimedFunctionManager.startTimedFunction(self.sendAliveRequests)
    ...

    The main process will call TimedFunctionManager.stopTimedFunctions() on exit or reload.
    This makes sure all thread get terminated.
    """

    timed_function_handlers = []

    @staticmethod
    def startTimedFunction(timed_function, *args, **kwargs):
        """
        Start a timed function and keep track of all running functions.
        """
        handler = timed_function(*args, **kwargs)
        TimedFunctionManager.timed_function_handlers.append(handler)
        return handler

    @staticmethod
    def stopTimedFunctions(handler=False):
        """
        Stop all timed functions. They are started as daemon, so when a reaload occurs, they will not finish cause the
        main thread still is running. This takes care of this issue.
        """
        if not TimedFunctionManager.timed_function_handlers:
            return
        # Clear provided handler only.
        if handler and handler in TimedFunctionManager.timed_function_handlers:
            handler.set()
            TimedFunctionManager.timed_function_handlers.remove(handler)
            return
        # Clear all timed functions
        for handler in TimedFunctionManager.timed_function_handlers:
            handler.set()
            TimedFunctionManager.timed_function_handlers.remove(handler)
        TimedFunctionManager.timed_function_handlers = []

def restartMainProcess():
    """
    Reload the whole LumberMill process. This code is a direct copy from tornado-3.1.1-py2.7.egg/tornado/autoreload.py.
    """
    if hasattr(signal, "setitimer"):
        # Clear the alarm signal set by
        # ioloop.set_blocking_log_threshold so it doesn't fire
        # after the exec.
        signal.setitimer(signal.ITIMER_REAL, 0, 0)
    # sys.path fixes: see comments at top of file.  If sys.path[0] is an empty
    # string, we were (probably) invoked with -m and the effective path
    # is about to change on re-exec.  Add the current directory to $PYTHONPATH
    # to ensure that the new process sees the same path we did.
    path_prefix = '.' + os.pathsep
    if sys.path[0] == '' and not os.environ.get("PYTHONPATH", "").startswith(path_prefix):
        os.environ["PYTHONPATH"] = (path_prefix + os.environ.get("PYTHONPATH", ""))
    if sys.platform == 'win32':
        # os.execv is broken on Windows and can't properly parse command line
        # arguments and executable name if they contain whitespaces. subprocess
        # fixes that behavior.
        subprocess.Popen([sys.executable] + sys.argv)
        sys.exit(0)
    else:
        try:
            os.execv(sys.executable, [sys.executable] + sys.argv)
        except OSError:
            # Mac OS X versions prior to 10.6 do not support execv in
            # a process that contains multiple threads.  Instead of
            # re-executing in the current process, start a new one
            # and cause the current process to exit.  This isn't
            # ideal since the new process is detached from the parent
            # terminal and thus cannot easily be killed with ctrl-C,
            # but it's better than not being able to autoreload at
            # all.
            # Unfortunately the errno returned in this case does not
            # appear to be consistent, so we can't easily check for
            # this error specifically.
            os.spawnv(os.P_NOWAIT, sys.executable,
                      [sys.executable] + sys.argv)
            sys.exit(0)

class AnsiColors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    LIGHTBLUE = '\033[34m'
    YELLOW = '\033[33m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'

def coloredConsoleLogging(fn):
    # add methods we need to the class
    def new(*args):
        levelno = args[1].levelno
        if(levelno>=50):
            color = AnsiColors.FAIL
        elif(levelno>=40):
            color = AnsiColors.FAIL
        elif(levelno>=30):
            color = AnsiColors.WARNING
        elif(levelno>=20):
            color = AnsiColors.LIGHTBLUE
        elif(levelno>=10):
            color = AnsiColors.OKGREEN
        else:
            color = AnsiColors.LIGHTBLUE
        args[1].msg = color + args[1].msg +  AnsiColors.ENDC # normal
        return fn(*args)
    return new

class Node:
    def __init__(self, module):
        self.children = []
        self.module = module

    def addChild(self, node):
        self.children.append(node)

def hasLoop(node, stack=[]):
    if not stack:
        stack.append(node)
    for current_node in node.children:
        if current_node in stack:
            return [current_node]
        stack.append(current_node)
        return hasLoop(current_node, stack)
    return []


# -*- coding: utf-8 -*-
import sys

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseModule import BaseModule
from lumbermill.utils.Decorators import ModuleDocstringParser
from lumbermill.utils.RedisAsyncClient import AsyncRedisClient


@ModuleDocstringParser
class RedisChannel(BaseModule):
    """
    Subscribes to a redis channels and passes incoming events to receivers.

    channel: Name of redis channel to subscribe to.
    server: Redis server to connect to.
    port: Port redis server is listening on.
    db: Redis db.
    password: Redis password.

    Configuration template:

    - RedisChannel:
       channel:                         # <type: string; is: required>
       server:                          # <default: 'localhost'; type: string; is: optional>
       port:                            # <default: 6379; type: integer; is: optional>
       db:                              # <default: 0; type: integer; is: optional>
       password:                        # <default: None; type: None||string; is: optional>
       receivers:
        - NextModule
    """

    module_type = "input"
    """Set module type"""
    can_run_forked = False

    def configure(self, configuration):
         # Call parent configure method
        BaseModule.configure(self, configuration)
        try:
            self.client = AsyncRedisClient(address=(self.getConfigurationValue('server'), self.getConfigurationValue('port')))
            if self.getConfigurationValue('db') != 0:
                self.client.fetch(('select', self.getConfigurationValue('db')), self.checkReply)
            if self.getConfigurationValue('password'):
                self.client.fetch(('select', self.getConfigurationValue('password')), self.checkReply)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not connect to redis store at %s. Exception: %s, Error: %s." % (self.getConfigurationValue('server'), etype, evalue))
            self.lumbermill.shutDown()
        try:
            self.client.fetch(('subscribe', self.getConfigurationValue('channel')), self.receiveEvent)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not subscribe to channel at redis store at %s. Exception: %s, Error: %s." % (self.getConfigurationValue('server'), etype, evalue))
        #autoreload.add_reload_hook(self.shutDown)
        #ioloop = IOLoop.instance()
        #ioloop.make_current()
        #try:
        #    ioloop.start()
        #except ValueError:
            # Ignore errors like "ValueError: I/O operation on closed kqueue fd". These might be thrown during a reload.
        #    pass

    def checkReply(self, answer):
        if answer != "OK":
            self.logger.error("Could not connect to server %s. Server said: %s." % (self.getConfigurationValue('server'), answer))
            self.lumbermill.shutDown()

    def handleEvent(self, event):
        if event[0] != 'message':
            return
        yield DictUtils.getDefaultEventDict(dict={"received_from": '%s' % event[1], "data": event[2]}, caller_class_name=self.__class__.__name__)


# -*- coding: utf-8 -*-
import zmq
import sys
import time
import threading
import Queue

def usage():
    sys.stdout = sys.stderr
    print('Usage: spam_zmq.py -m [push|pub] -c count host [port 5151]')
    print('When using pub the topic will be set to "LoadTest"')
    sys.exit(2)

if len(sys.argv) < 5:
    usage()
    sys.exit()

mode = sys.argv[2]
count = int(eval(sys.argv[4]))
host = sys.argv[5]
port = int(sys.argv[6]) if len(sys.argv) == 7 else 5151

if mode not in ['push', 'pub'] or type(count) is not int or type(port) is not int:
    usage()

class Worker(threading.Thread):

    def __init__(self, queue):
        threading.Thread.__init__(self)
        self.daemon = True
        self.queue = queue
        self.connected = False

    def run(self):
        context = zmq.Context()
        if mode == 'push':
            sock = context.socket(zmq.PUSH)
        else:
            sock = context.socket(zmq.PUB)
        try:
            sock.connect('tcp://%s:%s' % (host, port))
            self.connected = True
        except:
            etype, evalue, etb = sys.exc_info()
            print("Failed to connect to %s:%s. Exception: %s, Error: %s." % (host, port, etype, evalue))
            self.connected = False
        while self.connected:
            try:
                message = self.queue.get(timeout=.5)
                if mode == 'push':
                    sock.send(message)
                else:
                    sock.send('LoadTest %s' % message)
            except Queue.Empty:
                self.connected = False
                break
        sock.close()
        return

class ZmqLoadTester():

    def __init__(self, num_workers=10):
        self.lock = threading.Lock()
        self.counter = 0
        self.queue = Queue.Queue(10)
        self.num_workers = num_workers

    def start(self, message):
        total_item_count = count
        workers = set()
        for i in xrange(0, self.num_workers):
            worker = Worker(self.queue)
            worker.start()
            workers.add(worker)
            time.sleep(.1)
            if not worker.connected:
                sys.exit()
        print("Start load test.")
        start = time.time()
        for counter in xrange(0, total_item_count):
            if counter % 1000 == 0:
                sys.stdout.write("Message already sent: %s. Mean req/s: %s\r" % (counter+1, counter / (time.time()-start)))
                sys.stdout.flush()
            now = "%f" % time.time()
            self.queue.put(message % now)
        for worker in workers:
            worker.join()
        stop = time.time()
        print("Message sent: %s. Took %s. Mean req/s: %s" % (counter+1, stop-start, total_item_count / (stop-start)))

if __name__ == '__main__':
    lt = ZmqLoadTester(5)
    lt.start("<13>229.25.18.182 - - [28/Jul/2006:10:27:10 -0300] \"GET /cgi-bin/try/9153/?param1=Test&param2=%s HTTP/1.0\" 200 3395 \"Mozilla/5.0 (Linux; U; Android 2.3.5; en-in; HTC_DesireS_S510e Build/GRJ90) AppleWebKit/533.1 (KHTML, like Gecko) Version/4.0 Mobile Safari/533.1\"")


# -*- coding: utf-8 -*-
import re
import sys

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser, setInterval
from lumbermill.utils.misc import TimedFunctionManager


@ModuleDocstringParser
class Math(BaseThreadedModule):
    """
    Execute arbitrary math functions.

    Simple example to cast nginx request time (seconds with milliseconds as float) to apache request time
    (microseconds as int):

    - Math:
        filter: if $(server_type) == "nginx"
        target_field: request_time
        function: int(float($(request_time)) * 1000)

    If interval is set, the results of <function> will be collected for the interval time and the final result
    will be calculated via the <results_function>.

    function: the function to be applied to/with the event data.
    results_function: if interval is configured, use this function to calculate the final result.
    interval: Number of seconds to until.
    target_field: event field to store the result in.

    Configuration template:

    - Math:
       function:                        # <type: string; is: required>
       results_function:                # <default: None; type: None||string; is: optional if interval is None else required>
       interval:                        # <default: None; type: None||float||integer; is: optional>
       target_field:                    # <default: None; type: None||string; is: optional>
       receivers:
        - NextModule
    """

    module_type = "modifier"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.results = []
        function_str = "lambda event: " + re.sub('%\((.*?)\)s', r"event.get('\1', False)", self.getConfigurationValue('function'))
        self.function = self.compileFunction(function_str)
        if self.getConfigurationValue('results_function'):
            function_str = "lambda results: "+ re.sub('%\((.*?)\)s', r"results", self.getConfigurationValue('results_function'))
            self.results_function = self.compileFunction(function_str)
        self.target_field = self.getConfigurationValue('target_field')
        self.interval = self.getConfigurationValue('interval')

    def compileFunction(self, function_str):
        try:
            lambda_function = eval(function_str)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Failed to compile function: %s. Exception: %s, Error: %s." % (function_str, etype, evalue))
            self.lumbermill.shutDown()
        return lambda_function

    def getEvaluateFunc(self):
        @setInterval(self.interval)
        def timedEvaluateFacets():
            if self.results:
                self.evaluateResults()
        return timedEvaluateFacets

    def initAfterFork(self):
        if self.interval:
            self.evaluate_facet_data_func = self.getEvaluateFunc()
            self.timed_func_handler = TimedFunctionManager.startTimedFunction(self.evaluate_facet_data_func)
        BaseThreadedModule.initAfterFork(self)

    def evaluateResults(self):
        results = self.results
        self.results = []
        try:
            result = self.results_function(results)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Failed to evaluate result function %s. Exception: %s, Error: %s." % (self.getConfigurationValue('results_function'), etype, evalue))
            return
        if self.target_field:
            event_dict = {self.target_field: result}
        else:
            event_dict = {'math_result': result}
        event = DictUtils.getDefaultEventDict(event_dict, caller_class_name=self.__class__.__name__, event_type='math')
        self.sendEvent(event)

    def handleEvent(self, event):
        try:
            result = self.function(event)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Failed to evaluate function %s. Exception: %s, Error: %s." % (self.getConfigurationValue('function'), etype, evalue))
        if self.interval:
            self.results.append(result)
        elif self.target_field:
            event[self.target_field] = result
        yield event


# -*- coding: utf-8 -*-
import time

import datetime

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class AddDateTime(BaseThreadedModule):
    """
    Add a field with the current datetime.

    Configuration template:

    - AddDateTime:
       target_field:                    # <default: '@timestamp'; type: string; is: optional>
       format:                          # <default: '%Y-%m-%dT%H:%M:%S'; type: string; is: optional>
       receivers:
        - NextModule
    """
    module_type = "modifier"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.format = self.getConfigurationValue('format')
        self.target_field = self.getConfigurationValue('target_field')

    def handleEvent(self, event):
        event[self.target_field] = datetime.datetime.utcnow().strftime(self.format)
        yield event


# -*- coding: utf-8 -*-
import sys

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser

@ModuleDocstringParser
class ExecPython(BaseThreadedModule):
    """
    Execute python code.

    To make sure that the yaml parser keeps the tabs in the source code, ensure that the code is preceded by a comment.
    E.g.:

- ExecPython:
    source: |
      # Useless comment...
        try:
            imported = math
        except NameError:
            import math
        event['request_time'] = math.ceil(event['request_time'] * 1000)

    imports: Modules to import, e.g. re, math etc.
    code: Code to execute.
    debug: Set to True to output the string that will be executed.

    Configuration template:

    - ExecPython:
       imports:                         # <default: []; type: list; is: optional>
       source:                          # <type: string; is: required>
       debug:                           # <default: False; type: boolean; is: optional>
       receivers:
        - NextModule
    """

    module_type = "misc"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        for module in self.getConfigurationValue("imports"):
            exec("import %s" % module)
        source = "def UserFunc(event):\n%s" % self.getConfigurationValue("source")
        if self.getConfigurationValue("debug"):
            print source
        exec(source)
        try:
            pass
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("ExecPython failed while creating user function. Exception: %s, Error: %s" % (etype, evalue))
            self.lumbermill.shutDown()
        self._func = locals()["UserFunc"]


    def handleEvent(self, event):
        try:
            self._func(event)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.warning("ExecPython failed. Exception: %s, Error: %s" % (etype, evalue))
        yield event



# -*- coding: utf-8 -*-
import sys

from kafka import KafkaConsumer

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class Kafka(BaseThreadedModule):
    """
    Simple kafka input.


    Configuration template:

    - Kafka:
       brokers:                         # <type: list; is: required>
       topics:                          # <type: string||list; is: required>
       client_id:                       # <default: 'kafka.consumer.kafka'; type: string; is: optional>
       group_id:                        # <default: None; type: None||string; is: optional>
       fetch_message_max_bytes:         # <default: 1048576; type: integer; is: optional>
       fetch_min_bytes:                 # <default: 1; type: integer; is: optional>
       fetch_wait_max_ms:               # <default: 100; type: integer; is: optional>
       refresh_leader_backoff_ms:       # <default: 200; type: integer; is: optional>
       socket_timeout_ms:               # <default: 10000; type: integer; is: optional>
       auto_offset_reset:               # <default: 'largest'; type: string; is: optional>
       auto_commit_enable:              # <default: False; type: boolean; is: optional>
       auto_commit_interval_ms:         # <default: 60000; type: integer; is: optional>
       consumer_timeout_ms:             # <default: -1; type: integer; is: optional>
       receivers:
        - NextModule
    """

    module_type = "input"
    """Set module type"""
    can_run_forked = True

    def configure(self, configuration):
        # Call parent configure method.
        BaseThreadedModule.configure(self, configuration)
        self.auto_commit_enable = self.getConfigurationValue('auto_commit_enable')

    def initAfterFork(self):
        try:
            self.consumer = KafkaConsumer(self.getConfigurationValue('topics'),
                                          client_id=self.getConfigurationValue('client_id'),
                                          group_id=self.getConfigurationValue('group_id'),
                                          fetch_message_max_bytes=self.getConfigurationValue('fetch_message_max_bytes'),
                                          fetch_min_bytes=self.getConfigurationValue('fetch_min_bytes'),
                                          fetch_wait_max_ms=self.getConfigurationValue('fetch_wait_max_ms'),
                                          refresh_leader_backoff_ms=self.getConfigurationValue('refresh_leader_backoff_ms'),
                                          metadata_broker_list=self.getConfigurationValue('brokers'),
                                          socket_timeout_ms=self.getConfigurationValue('socket_timeout_ms'),
                                          auto_offset_reset=self.getConfigurationValue('auto_offset_reset'),
                                          auto_commit_enable=self.getConfigurationValue('auto_commit_enable'),
                                          auto_commit_interval_ms=self.getConfigurationValue('auto_commit_interval_ms'),
                                          consumer_timeout_ms=self.getConfigurationValue('consumer_timeout_ms')
                                          )
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not create kafka consumer. Exception: %s, Error: %s." % ( etype, evalue))
            self.lumbermill.shutDown()

    def run(self):
        while self.alive:
            for kafka_event in self.consumer:
                event = DictUtils.getDefaultEventDict(dict={"topic": kafka_event.topic, "data": kafka_event.value}, caller_class_name=self.__class__.__name__)
                self.sendEvent(event)
                if(self.auto_commit_enable):
                    self.consumer.task_done(kafka_event)



# -*- coding: utf-8 -*-
import time
from collections import defaultdict

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser, setInterval
from lumbermill.utils.DynamicValues import mapDynamicValue
from lumbermill.utils.misc import TimedFunctionManager


@ModuleDocstringParser
class Throttle(BaseThreadedModule):
    """
    Throttle event count over a given time period.

    key: Identifies events as being the "same". Dynamic notations can be used here.
    timeframe: Time window in seconds from first encountered event to last.
    min_count: Minimal count of same events to allow event to be passed on.
    max_mount: Maximum count of same events before same events will be blocked.
    backend: Name of a key::value store plugin. When running multiple instances of gp this backend can be used to synchronize events across multiple instances.
    backend_key_prefix: Prefix for the backend key.

    Configuration template:

    - Throttle:
       key:                             # <type:string; is: required>
       timeframe:                       # <default: 600; type: integer; is: optional>
       min_count:                       # <default: 1; type: integer; is: optional>
       max_count:                       # <default: 1; type: integer; is: optional>
       backend:                         # <default: None; type: None||string; is: optional>
       backend_key_prefix:              # <default: "lumbermill:throttle"; type: string; is: optional>
       receivers:
        - NextModule
    """

    module_type = "misc"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.throttled_events_info = defaultdict(int)
        self.key = self.getConfigurationValue('key')
        self.timeframe = self.getConfigurationValue('timeframe')
        self.min_count = self.getConfigurationValue('min_count')
        self.max_count = self.getConfigurationValue('max_count')
        self.backend_key_prefix = self.getConfigurationValue('backend_key_prefix')
        self.persistence_backend = None
        if self.getConfigurationValue('backend'):
            backend_info = self.lumbermill.getModuleInfoById(self.getConfigurationValue('backend'))
            if not backend_info:
                self.logger.error("Could not find %s backend for persistant storage." % (self.getConfigurationValue('backend')))
                self.lumbermill.shutDown()
                return
            self.persistence_backend = backend_info['instances'][0]

    def setAndGetEventCountByKey(self, key):
        now = time.time()
        if self.persistence_backend:
            throttled_event_info = self.persistence_backend.get("%s:%s" % (self.backend_key_prefix, key))
        else:
            try:
                throttled_event_info = self.throttled_events_info[key]
            except KeyError:
                throttled_event_info  = None
        if not throttled_event_info or (now - throttled_event_info['ctime']) > self.timeframe:
            throttled_event_info = {'count': 1, 'ctime': now}
        else:
            throttled_event_info['count'] += 1
        if self.persistence_backend:
            self.persistence_backend.set("%s:%s" % (self.backend_key_prefix, key), throttled_event_info)
        else:
            self.throttled_events_info[key] = throttled_event_info
        return throttled_event_info['count']

    def getGcThrottledEventsInfoFunc(self):
        @setInterval(15)
        def gcThrottledEventsInfo():
            now = time.time()
            if self.persistence_backend:
                throttled_event_keys = self.persistence_backend.client.keys("%s:*" % self.backend_key_prefix)
                if not throttled_event_keys:
                    return
                keys_to_delete = []
                for key in throttled_event_keys:
                    throttled_event_info = self.persistence_backend.get(key)
                    if (now - throttled_event_info['ctime']) > self.timeframe:
                        keys_to_delete.append(key)
                if keys_to_delete:
                    self.persistence_backend.client.delete(*keys_to_delete)
                return
            throttled_events_info_copy = self.throttled_events_info.copy()
            for key, throttled_event_info in throttled_events_info_copy.items():
                if (now - throttled_event_info['ctime']) > self.timeframe:
                    del self.throttled_events_info[key]
        return gcThrottledEventsInfo

    def initAfterFork(self):
        self.gc_throttled_events_info = self.getGcThrottledEventsInfoFunc()
        self.timed_func_handler = TimedFunctionManager.startTimedFunction(self.gc_throttled_events_info)
        BaseThreadedModule.initAfterFork(self)

    def handleEvent(self, event):
        throttled_event_key = mapDynamicValue(self.key, event)
        throttled_event_count = self.setAndGetEventCountByKey(throttled_event_key)
        if self.min_count <= throttled_event_count <= self.max_count:
            yield event


# -*- coding: utf-8 -*-
import cPickle
import sys

import redis

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Buffers import Buffer
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class KeyValueStore(BaseThreadedModule):
    """
    A simple wrapper around the python simplekv module.

    It can be used to store results of modules in all simplekv supported backends.

    When set, the following options cause RedisStore to use a buffer for setting values.
    Multiple values are set via the pipe command, which speeds up storage. Still this comes at a price.
    Buffered values, that have not yet been send to redis, will be lost when LumberMill crashes.

    backend: backends supported by [simplekv](http://pythonhosted.org//simplekv/)
    store_interval_in_secs: Sending data to redis in x seconds intervals.
    batch_size: Sending data to redis if count is above, even if store_interval_in_secs is not reached.
    backlog_size: Maximum count of values waiting for transmission. Values above count will be dropped.

    Configuration template:

    - KeyValueStore:
       backend:                         # <default: 'DictStore'; type: string; is: optional>
       server:                          # <default: None; type: None||string; is: required if backend in ['RedisStore', 'MemcacheStore'] and cluster is None else optional>
       cluster:                         # <default: None; type: None||dictionary; is: required if backend == 'RedisStore' and server is None else optional>
       port:                            # <default: 6379; type: integer; is: optional>
       db:                              # <default: 0; type: integer; is: optional>
       password:                        # <default: None; type: None||string; is: optional>
       socket_timeout:                  # <default: 10; type: integer; is: optional>
       charset:                         # <default: 'utf-8'; type: string; is: optional>
       errors:                          # <default: 'strict'; type: string; is: optional>
       decode_responses:                # <default: False; type: boolean; is: optional>
       unix_socket_path:                # <default: None; type: None||string; is: optional>
       batch_size:                      # <default: None; type: None||integer; is: optional>
       store_interval_in_secs:          # <default: None; type: None||integer; is: optional>
       backlog_size:                    # <default: 5000; type: integer; is: optional>
    """
    module_type = "stand_alone"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        backend = self.getConfigurationValue('backend')
        self.backend_client = None
        if backend == 'DictStore':
            import simplekv.memory
            self.backend_client = None
            self.kv_store = simplekv.memory.DictStore()
        elif backend == 'RedisStore':
            import simplekv.memory.redisstore
            self.backend_client = self.getRedisClient()
            self.kv_store = simplekv.memory.redisstore.RedisStore(self.backend_client)
        elif backend == 'MemcacheStore':
            import simplekv.memory.memcachestore
            self.backend_client = self.getMemcacheClient()
            self.kv_store = simplekv.memory.memcachestore.MemcacheStore(self.backend_client)

        self.set_buffer = None
        if self.getConfigurationValue('store_interval_in_secs') or self.getConfigurationValue('batch_size'):
            if backend == 'RedisStore':
                self.set_buffer = Buffer(self.getConfigurationValue('batch_size'), self.setRedisBufferedCallback, self.getConfigurationValue('store_interval_in_secs'), maxsize=self.getConfigurationValue('backlog_size'))
            else:
                 self.set_buffer = Buffer(self.getConfigurationValue('batch_size'), self.setBufferedCallback, self.getConfigurationValue('store_interval_in_secs'), maxsize=self.getConfigurationValue('backlog_size'))
            self._set = self.set
            self.set = self.setBuffered
            self._get = self.get
            self.get = self.getBuffered
            self._delete = self.delete
            self.delete = self.deleteBuffered
            self._pop = self.pop
            self.pop = self.popBuffered

    def getRedisClient(self):
        if not self.getConfigurationValue('cluster') or len(self.getConfigurationValue('cluster')) == 0:
            redis_store = self.getConfigurationValue('server')
            client = self.getSimpleRedisClient()
        else:
            redis_store = self.getConfigurationValue('cluster')
            client = self.getClusterRedisClient()
        try:
            client.ping()
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not connect to redis store at %s. Exception: %s, Error: %s." % (redis_store, etype, evalue))
            self.lumbermill.shutDown()
        return client

    def getMemcacheClient(self):
        client = None
        # TODO: implement memcache client
        return client

    def getSimpleRedisClient(self):
        try:
            client = redis.StrictRedis(host=self.getConfigurationValue('server'),
                                       port=self.getConfigurationValue('port'),
                                       db=self.getConfigurationValue('db'),
                                       password=self.getConfigurationValue('password'),
                                       socket_timeout=self.getConfigurationValue('socket_timeout'),
                                       charset=self.getConfigurationValue('charset'),
                                       errors=self.getConfigurationValue('errors'),
                                       decode_responses=self.getConfigurationValue('decode_responses'),
                                       unix_socket_path=self.getConfigurationValue('unix_socket_path'))
            return client
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not connect to redis store at %s. Exception: %s, Error: %s." % (self.getConfigurationValue['server'], etype, evalue))
            self.lumbermill.shutDown()

    def getClusterRedisClient(self):
        try:
            import rediscluster
        except ImportError:
            self.logger.error("Could not import rediscluster module. To install follow instructions @https://github.com/salimane/rediscluster-py")
            self.lumbermill.shutDown()
        # TODO: Implement a locking mechanism for the cluster client.
        # Some modules like Facet depend on this.
        cluster = {'nodes': {}, 'master_of': {}}
        counter = 1
        for master_node, slave_nodes in self.getConfigurationValue('cluster').items():
            master_node_key = "node_%d" % counter
            node_name_or_ip, node_port = self._parseRedisServerAddress(master_node)
            cluster['nodes'].update({master_node_key: {'host': node_name_or_ip, 'port': node_port}})
            if 'default_node' not in cluster:
                cluster['default_node'] = master_node
            if type(slave_nodes) is str:
                slave_nodes = [slave_nodes]
            for slave_node in slave_nodes:
                counter += 1
                slave_node_key = "node_%d" % counter
                node_name_or_ip, node_port = self._parseRedisServerAddress(slave_node)
                cluster['nodes'].update({slave_node_key: {'host':node_name_or_ip, 'port': node_port}})
                cluster['master_of'].update({master_node_key: slave_node_key})
        try:
            client = rediscluster.StrictRedisCluster(cluster=cluster, db=self.getConfigurationValue('db'))
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not connect to redis store at %s. Exception: %s, Error: %s." % (self.getConfigurationValue['cluster'], etype, evalue))
            self.lumbermill.shutDown()
        return client

    def _parseRedisServerAddress(self, node_address):
        try:
            node_name_or_ip, node_port = node_address.split(":")
        except ValueError:
            node_name_or_ip = node_address
            node_port = self.getConfigurationValue('port')
        return (node_name_or_ip, node_port)

    def iterKeys(self):
        for key in self.kv_store.iter_keys():
            yield key

    def getClient(self):
        return self.backend_client

    def getLock(self, name, timeout=None, sleep=0.1):
        lock = False
        try:
            lock = self.backend_client.lock(name, timeout, sleep)
        except:
            pass
        return lock

    def set(self, key, value, ttl=0, pickle=True):
        if pickle is True:
            try:
                value = cPickle.dumps(value)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error("Could not store %s:%s in redis. Exception: %s, Error: %s." % (key, value, etype, evalue))
                raise
        if ttl:
            self.kv_store.put(key, value, ttl_secs=ttl)
        else:
            self.kv_store.put(key, value)

    def setBuffered(self, key, value, ttl=0, pickle=True):
        if pickle is True:
            try:
                value = cPickle.dumps(value)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error("Could not store %s:%s in redis. Exception: %s, Error: %s." % (key, value, etype, evalue))
                raise
        if ttl:
            self.set_buffer.append({'key':key, 'ttl': ttl, 'value': value})
        else:
            self.set_buffer.append({'key':key, 'value': value})

    def setBufferedCallback(self, values):
        for value in values:
            if 'ttl' in value:
                self._set(value['key'], value['value'], value['ttl'])
            else:
                self._set(value['key'], value['value'])

    def setRedisBufferedCallback(self, values):
        pipe = self.backend_client.pipeline()
        for value in values:
            if 'ttl' in value:
                pipe.setex(value['key'], value['ttl'], value['value'])
            else:
                pipe.set(value['key'], value['value'])
        try:
            pipe.execute()
            return True
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not flush buffer. Exception: %s, Error: %s." % (etype, evalue))

    def get(self, key, unpickle=True):
        value = self.kv_store.get(key)
        if unpickle and value:
            try:
                value = cPickle.loads(value)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error("Could not unpickle %s:%s from redis. Exception: %s, Error: %s." % (key, value, etype, evalue))
                raise
        return value

    def getBuffered(self, key, unpickle=True):
        try:
            value_idx = next(index for (index, entry) in enumerate(self.set_buffer.buffer) if entry["key"] == key)
            return self.set_buffer.buffer[value_idx]
        except:
            return self._get(key, unpickle)

    def delete(self, key):
        self.kv_store.delete(key)

    def deleteBuffered(self, key):
        try:
            value_idx = next(index for (index, entry) in enumerate(self.set_buffer.buffer) if entry["key"] == key)
            self.set_buffer.buffer.pop(value_idx)
            return
        except:
            self._delete(key)

    def pop(self, key, unpickle=True):
        value = self.get(key, unpickle)
        if value:
            self.delete(key)
        return value

    def popBuffered(self, key, unpickle=True):
        try:
            value_idx = next(index for (index, entry) in enumerate(self.set_buffer.buffer) if entry["key"] == key)
            return self.set_buffer.buffer.pop(value_idx)
        except:
            return self._pop(key, unpickle)

    def shutDown(self):
        try:
            self.buffer.flush()
        except:
            pass
        BaseThreadedModule.shutDown(self)



# -*- coding: utf-8 -*-
import sys
import time

import redis

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class RedisList(BaseThreadedModule):
    """
    Subscribes to a redis channels/lists and passes incoming events to receivers.

    lists: Name of redis lists to subscribe to.
    server: Redis server to connect to.
    port: Port redis server is listening on.
    batch_size: Number of events to return from redis list.
    db: Redis db.
    password: Redis password.
    timeout: Timeout in seconds.

    Configuration template:

    - RedisList:
       lists:                           # <type: string||list; is: required>
       server:                          # <default: 'localhost'; type: string; is: optional>
       port:                            # <default: 6379; type: integer; is: optional>
       batch_size:                      # <default: 1; type: integer; is: optional>
       db:                              # <default: 0; type: integer; is: optional>
       password:                        # <default: None; type: None||string; is: optional>
       timeout:                         # <default: 0; type: integer; is: optional>
       receivers:
        - NextModule
    """

    module_type = "input"
    """Set module type"""
    can_run_forked = True

    def configure(self, configuration):
         # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.lists = self.getConfigurationValue('lists')
        if not isinstance(self.lists, list):
            self.lists = [self.lists]
        self.timeout = self.getConfigurationValue('timeout')
        self.batch_size = self.getConfigurationValue('batch_size')
        self.client = redis.StrictRedis(host=self.getConfigurationValue('server'),
                                        port=self.getConfigurationValue('port'),
                                        password=self.getConfigurationValue('password'),
                                        db=self.getConfigurationValue('db'))
        try:
            self.client.ping()
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not connect to redis store at %s. Exception: %s, Error: %s." % (self.getConfigurationValue('server'), etype, evalue))
            self.lumbermill.shutDown()
        # Monkeypatch run method to use the correct handle event method.
        if self.batch_size == 1:
            self.run = self.handleSingleEvent
        else:
            self.run = self.handleBatchEvents

    def run(self):
        self.logger.error("Monkeypatching the run method of RedisList seems to have failed.")
        self.lumbermill.shutDown()

    def handleSingleEvent(self):
        while self.alive:
            event = None
            try:
                event = self.client.blpop(self.lists, timeout=self.timeout)
            except:
                exc_type, exc_value, exc_tb = sys.exc_info()
                self.logger.error("Could not read data from redis list(s) %s. Exception: %s, Error: %s." % (self.lists, exc_type, exc_value))
            if event:
                event = DictUtils.getDefaultEventDict(dict={"received_from": '%s' % event[0], "data": event[1]}, caller_class_name=self.__class__.__name__)
                self.sendEvent(event)
            else:
                # Queue is exhausted. Sleep a bit and retry.
                time.sleep(.5)
                continue

    def handleBatchEvents(self):
        pipeline = self.client.pipeline()
        while self.alive:
            for _ in range(0, self.batch_size):
                pipeline.blpop(self.lists, timeout=self.timeout)
            try:
                events = pipeline.execute()
            except:
                exc_type, exc_value, exc_tb = sys.exc_info()
                self.logger.error("Could not read data from redis list(s) %s. Exception: %s, Error: %s." % (self.lists, exc_type, exc_value))
                continue
            for event in events:
                # If batch_size is bigger than events waiting in redis queue, the remaining entries will be filled with None values.
                # So break out if a None value is found.
                if not event:
                    # Queue is exhausted. Sleep a bit and retry.
                    time.sleep(.5)
                    break
                event = DictUtils.getDefaultEventDict(dict={"received_from": '%s' % event[0], "data": event[1]}, caller_class_name=self.__class__.__name__)
                self.sendEvent(event)


# -*- coding: utf-8 -*-
import collections
import os
import sys
import time
from cStringIO import StringIO

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Buffers import Buffer
from lumbermill.utils.Decorators import ModuleDocstringParser, setInterval
from lumbermill.utils.DynamicValues import mapDynamicValue
from lumbermill.utils.misc import TimedFunctionManager


@ModuleDocstringParser
class FileSink(BaseThreadedModule):
    """
    Store all received events in a file.

    file_name: absolute path to filen. String my contain pythons strtime directives and event fields, e.g. %Y-%m-%d.
    format: Which event fields to use in the logline, e.g. '$(@timestamp) - $(url) - $(country_code)'
    store_interval_in_secs: sending data to es in x seconds intervals.
    batch_size: sending data to es if event count is above, even if store_interval_in_secs is not reached.
    backlog_size: maximum count of events waiting for transmission. Events above count will be dropped.
    compress: Compress output as gzip or snappy file. For this to be effective, the chunk size should not be too small.

    Configuration template:

    - FileSink:
       file_name:                       # <type: string; is: required>
       format:                          # <default: '$(data)'; type: string; is: optional>
       store_interval_in_secs:          # <default: 10; type: integer; is: optional>
       batch_size:                      # <default: 500; type: integer; is: optional>
       backlog_size:                    # <default: 500; type: integer; is: optional>
       compress:                        # <default: None; type: None||string; values: [None,'gzip','snappy']; is: optional>
    """

    module_type = "output"
    """Set module type"""
    can_run_forked = False

    def configure(self, configuration):
         # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.batch_size = self.getConfigurationValue('batch_size')
        self.backlog_size = self.getConfigurationValue('backlog_size')
        self.file_name = self.getConfigurationValue('file_name')
        self.format = self.getConfigurationValue('format')
        self.compress = self.getConfigurationValue('compress')
        self.file_handles = {}
        if self.compress == 'gzip':
            try:
                # Import module into namespace of object. Otherwise it will not be accessible when process was forked.
                self.gzip_module = __import__('gzip')
            except ImportError:
                self.logger.error('Gzip compression selected but gzip module could not be loaded.')
                self.lumbermill.shutDown()
        if self.compress == 'snappy':
            try:
                self.snappy_module = __import__('snappy')
            except ImportError:
                self.logger.error('Snappy compression selected but snappy module could not be loaded.')
                self.lumbermill.shutDown()
        self.buffer = Buffer(self.batch_size, self.storeData, self.getConfigurationValue('store_interval_in_secs'), maxsize=self.backlog_size)
        TimedFunctionManager.startTimedFunction(self.closeStaleFileHandles)

    def getStartMessage(self):
        return "File: %s. Max buffer size: %d" % (self.file_name, self.getConfigurationValue('backlog_size'))

    def initAfterFork(self):
        BaseThreadedModule.initAfterFork(self)
        # As the buffer uses a threaded timed function to flush its buffer and thread will not survive a fork, init buffer here.
        self.buffer = Buffer(self.getConfigurationValue('batch_size'), self.storeData, self.getConfigurationValue('store_interval_in_secs'), maxsize=self.getConfigurationValue('backlog_size'))
        BaseThreadedModule.initAfterFork(self)

    @setInterval(60)
    def closeStaleFileHandles(self):
        """
        Close and delete file handles that are unused since 5 minutes.
        """
        for path, file_handle_data in self.file_handles.items():
            last_used_time_ago = time.time() - file_handle_data['lru']
            if last_used_time_ago < 300:
                continue
            self.logger.info('Closing stale file handle for %s.' % (path))
            file_handle_data['handle'].close()
            self.file_handles.pop(path)

    def closeAllFileHandles(self):
        for path, file_handle_data in self.file_handles.items():
            self.logger.info('Closing file handle for %s.' % path)
            file_handle_data['handle'].close()
            self.file_handles.pop(path)

    def ensurePathExists(self, path):
        dirpath = os.path.dirname(path)
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

    def handleEvent(self, event):
        self.buffer.append(event)
        yield None

    def getOrCreateFileHandle(self, path, mode):
        file_handle = None
        try:
            file_handle = self.file_handles[path]['handle']
            self.file_handles[path]['lru'] = time.time()
        except KeyError:
            try:
                file_handle = open(path, mode)
                self.file_handles[path] = {'handle': file_handle, 'lru': time.time()}
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error('Could no open %s for writing. Exception: %s, Error: %s.' % (path, etype, evalue))
        return file_handle

    def storeData(self, events):
        write_data = collections.defaultdict(str)
        for event in events:
            path = mapDynamicValue(self.file_name, mapping_dict=event, use_strftime=True)
            line = mapDynamicValue(self.format, mapping_dict=event)
            write_data["%s" % path] += line + "\n"
        for path, lines in write_data.items():
            try:
                self.ensurePathExists(path)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error('Could no create path %s. Events could not be written. Exception: %s, Error: %s.' % (path, etype, evalue))
                return
            mode = "a+"
            if self.compress == 'gzip':
                path += ".gz"
                mode += "b"
                lines = self.compressGzip(lines)
            elif self.compress == 'snappy':
                path += ".snappy"
                lines = self.compressSnappy(lines)
                mode += "b"
            try:
                fh = self.getOrCreateFileHandle(path, mode)
                fh.write(lines)
                return True
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error('Could no write event data to %s. Exception: %s, Error: %s.' % (path, etype, evalue))

    def shutDown(self):
        self.buffer.flush()
        self.closeAllFileHandles()
        BaseThreadedModule.shutDown(self)

    def compressGzip(self, data):
        buffer = StringIO()
        compressor = self.gzip_module.GzipFile(mode='wb', fileobj=buffer)
        try:
            compressor.write(data)
        finally:
            compressor.close()
        return buffer.getvalue()

    def compressSnappy(self, data):
        return self.snappy_module.compress(data)



# -*- coding: utf-8 -*-
import gzip
import sys
import types
import zlib
from cStringIO import StringIO

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class InflateParser(BaseThreadedModule):
    """
    Inflate any field with supported compression codecs.

    It will take the source fields and decompress them with the configured codecs. At the moment only gzip an zlib are
    supported.

    source_fields: Single field or list of fields to decompress.
    target_fields: Single field or list of fields to fill with decompressed data.
                   If not provided, contents of source_fields will be replaced.
    compression:   Compression lib to use for decompression.

    Configuration template:

    - InflateParser:
       source_fields:                   # <default: 'data'; type: string||list; is: optional>
       target_fields:                   # <default: None; type: None||string||list; is: optional>
       compression:                     # <default: 'gzip'; type: string; is: optional; values: ['gzip', 'zlib']>
       receivers:
        - NextModule
    """

    module_type = "parser"
    """Set module type"""
    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.source_fields = self.getConfigurationValue('source_fields')
        # Allow single string as well.
        if isinstance(self.source_fields, types.StringTypes):
            self.source_fields = [self.source_fields]
        self.target_fields = self.getConfigurationValue('target_fields')
        # Allow single string as well.
        if isinstance(self.target_fields, types.StringTypes):
            self.target_fields = [self.target_fields]
        # If target_fields were provided they need to be of same length as source_fields.
        if self.target_fields and len(self.source_fields) != len(self.target_fields):
            self.logger.error("Count of configured source and target fields need to match. Please check your configuration.")
            self.lumbermill.shutDown()
        self.compression = self.getConfigurationValue('compression')
        # Get compression specific handler.
        try:
            self.deflater = getattr(self, "deflate_with_%s" % self.compression)
        except AttributeError:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Unknown decompression lib: %s. Exception: %s, Error: %s" % (self.compression, etype, evalue))
            self.lumbermill.shutDown()

    def handleEvent(self, event):
        for idx, source_field in enumerate(self.source_fields):
            infalted_data = self.deflater(event[source_field])
            if not self.target_fields:
                event[source_field] = infalted_data
            else:
                event[self.target_fields[idx]] = infalted_data
        yield event

    def deflate_with_gzip(self, value):
        buffer = StringIO(value)
        try:
            deflated_value = gzip.GzipFile(fileobj=buffer)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.warning("Could not decompress data %s. Exception: %s, Error: %s." % (value, etype, evalue))
        return deflated_value.read()

    def deflate_with_zlib(self, value):
        try:
            # Adding 32 to windowBits will trigger header detection.
            deflated_value = zlib.decompress(value, zlib.MAX_WBITS|32)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.warning("Could not decompress data %s. Exception: %s, Error: %s." % (value, etype, evalue))
        return deflated_value



# -*- coding: utf-8 -*-
import collections
import os
import re
import sys

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Buffers import Buffer
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class MergeEvent(BaseThreadedModule):
    """
    Merge multiple event into a single one.

    In most cases, inputs will split an incoming stream at some kind of delimiter to produce events.
    Sometimes, the delimiter also occurs in the event data itself and splitting here is not desired.
    To mitigate this problem, this module can merge these fragmented events based on some configurable rules.

    Each incoming event will be buffered in a queue identified by <buffer_key>.
    If a new event arrives and <pattern> does not match for this event, the event will be appended to the buffer.
    If a new event arrives and <pattern> matches for this event, the buffer will be flushed prior to appending the event.
    After <flush_interval_in_secs> the buffer will also be flushed.
    Flushing the buffer will concatenate all contained event data to form one single new event.

    buffer_key: key to distinguish between different input streams

    buffer_key: A key to correctly group events.
    buffer_size: Maximum size of events in buffer. If size is exceeded a flush will be executed.
    flush_interval_in_secs: If interval is reached, buffer will be flushed.
    pattern: Pattern to match new events. If pattern matches, a flush will be executed prior to appending the event to buffer.
    glue: Join event data with glue as separator.

    Configuration template:

    - MergeEvent:
       buffer_key:                      # <default: "$(lumbermill.received_from)"; type: string; is: optional>
       buffer_size:                     # <default: 100; type: integer; is: optional>
       flush_interval_in_secs:          # <default: 1; type: None||integer; is: required if pattern is None else optional>
       pattern:                         # <default: None; type: None||string; is: required if flush_interval_in_secs is None else optional>
       match_field:                     # <default: "data"; type: string; is: optional>
       glue:                            # <default: ""; type: string; is: optional>
       receivers:
        - NextModule
    """

    module_type = "modifier"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.logstash_patterns = {}
        self.readLogstashPatterns()
        self.pattern = self.getConfigurationValue('pattern')
        if self.pattern:
            try:
                self.pattern = self.replaceLogstashPatterns(self.pattern)
                self.pattern = re.compile(self.pattern)
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.error("RegEx error for pattern %s. Exception: %s, Error: %s." % (self.pattern, etype, evalue))
                self.lumbermill.shutDown()
        self.match_field = self.getConfigurationValue('match_field')
        self.buffer_size = self.getConfigurationValue('buffer_size')
        self.flush_interval_in_secs = self.getConfigurationValue('flush_interval_in_secs')
        self.glue = self.getConfigurationValue('glue')

    def readLogstashPatterns(self):
        path = "%s/../assets/grok_patterns" % os.path.dirname(os.path.realpath(__file__))
        for (dirpath, dirnames, filenames) in os.walk(path):
            for filename in filenames:
                lines = [line.strip() for line in open('%s%s%s' % (dirpath, os.sep, filename))]
                for line_no, line in enumerate(lines):
                    if line == "" or line.startswith('#'):
                        continue
                    try:
                        pattern_name, pattern = line.split(' ', 1)
                        self.logstash_patterns[pattern_name] = pattern
                    except:
                        etype, evalue, etb = sys.exc_info()
                        self.logger.warning("Could not read logstash pattern in file %s%s%s, line %s. Exception: %s, Error: %s." % (dirpath,  os.sep, filename, line_no+1, etype, evalue))

    def replaceLogstashPatterns(self, regex_pattern):
        pattern_name_re = re.compile('%\{(.*?)\}')
        for match in pattern_name_re.finditer(regex_pattern):
            for pattern_name in match.groups():
                try:
                    logstash_pattern = self.replaceLogstashPatterns(self.logstash_patterns[pattern_name])
                    regex_pattern = regex_pattern.replace('%%{%s}' % pattern_name, logstash_pattern)
                except KeyError:
                    self.logger.warning("Could not parse logstash pattern %s. Pattern name not found in pattern files." % (pattern_name))
                    continue
        return regex_pattern

    def initAfterFork(self):
        # As the buffer uses a threaded timed function to flush its buffer and thread will not survive a fork, init buffer here.
        self.buffers = collections.defaultdict(lambda: Buffer(flush_size=self.buffer_size,
                                                                    callback=self.sendMergedEvent,
                                                                    interval=self.flush_interval_in_secs,
                                                                    maxsize=self.buffer_size))
        BaseThreadedModule.initAfterFork(self)

    def handleEvent(self, event):
        key = self.getConfigurationValue("buffer_key", event)
        # No pattern was defined, to merging of event data will only be based on the buffer key.
        if not self.pattern:
            self.buffers[key].append(event)
            yield None
            return
        # If pattern matches create new buffer with current key.
        if self.pattern.search(event[self.match_field]):
            # If a buffer with current key exists, flush it before appending again.
            if self.buffers[key].bufsize() > 0:
                self.buffers[key].flush()
            self.buffers[key].append(event)
            yield None
        # Append to existing buffer.
        elif self.buffers[key].bufsize() > 0:
            self.buffers[key].append(event)
            yield None
        # No buffer exists, so just pass this event on to next module.
        else:
            yield event

    def sendMergedEvent(self, events):
        if len(events) == 1:
            self.sendEvent(events[0])
            return True
        else:
            parent_event = events[0]
            parent_event['data'] = self.glue.join([event["data"] for event in events])
            caller_class_name = parent_event["lumbermill"].get("source_module", None)
            received_from = parent_event["lumbermill"].get("received_from", None)
            merged_event = DictUtils.getDefaultEventDict(parent_event, caller_class_name=caller_class_name, received_from=received_from)
            self.sendEvent(merged_event)
            return True


# -*- coding: utf-8 -*-
import types

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class LineParser(BaseThreadedModule):
    r"""
    Line parser.

    Decode:
    Will split the data in source fields and emit parts as new events. So if e.g. data field contains:
    message-a|message-b|message-c
    you can split this field by "|" and three new events will be created with message-a, message-b and message-c as
    payload.

    The original event will be discarded.

    source_fields:  Input fields to split. Can be a single field or a list of fields.
    seperator:      Char used as line seperator.
    target_field:   event field to be filled with the new data.

    Configuration template:

    - LineParser:
       source_fields:                   # <default: 'data'; type: string||list; is: optional>
       seperator:                       # <default: '\n'; type: string; is: optional>
       target_field:                    # <default: 'data'; type:string; is: optional>
       keep_original:                   # <default: False; type: boolean; is: optional>
       receivers:
        - NextModule
    """

    module_type = "parser"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.source_fields = self.getConfigurationValue('source_fields')
        # Allow single string as well.
        if isinstance(self.source_fields, types.StringTypes):
            self.source_fields = [self.source_fields]
        self.seperator = self.getConfigurationValue('seperator')
        self.target_field = self.getConfigurationValue('target_field')
        self.drop_original = not self.getConfigurationValue('keep_original')

    def handleEvent(self, event):
        for source_field in self.source_fields:
            if source_field not in event:
                continue
            decoded_datasets = event[source_field].split(self.seperator)
            copy_event = False
            for decoded_data in decoded_datasets:
                if copy_event:
                    event = event.copy()
                copy_event = True
                if self.drop_original:
                    event.pop(source_field, None)
                event.update({self.target_field: decoded_data})
                yield event




# -*- coding: utf-8 -*-
import os

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser, setInterval
from lumbermill.utils.StatisticCollector import StatisticCollector, MultiProcessStatisticCollector
from lumbermill.utils.misc import AnsiColors, TimedFunctionManager


@ModuleDocstringParser
class SimpleStats(BaseThreadedModule):
    """
    Collect and log some simple lumbermill statistic data.

    Use this module if you just need some simple statistics on how many events are passing through lumbermill.
    Per default, statistics will just be send to stdout.

    As a side note: This module inits MultiProcessStatisticCollector. As it uses multiprocessing.Manager().dict()
    this will start another process. So if you use SimpleStats, you will see workers + 1 processes in the process
    list.

    Configuration template:

    - SimpleStats:
       interval:                        # <default: 10; type: integer; is: optional>
       event_type_statistics:           # <default: True; type: boolean; is: optional>
       receive_rate_statistics:         # <default: True; type: boolean; is: optional>
       waiting_event_statistics:        # <default: False; type: boolean; is: optional>
       emit_as_event:                   # <default: False; type: boolean; is: optional>
    """

    module_type = "misc"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.emit_as_event = self.getConfigurationValue('emit_as_event')
        self.interval = self.getConfigurationValue('interval')
        self.event_type_statistics = self.getConfigurationValue('event_type_statistics')
        self.stats_collector = StatisticCollector()
        self.mp_stats_collector = MultiProcessStatisticCollector()
        self.module_queues = {}

    def getRunTimedFunctionsFunc(self):
        @setInterval(self.interval)
        def runTimedFunctionsFunc():
            self.accumulateReceiveRateStats()
            self.accumulateEventTypeStats()
            if self.lumbermill.is_master():
                self.printIntervalStatistics()
        return runTimedFunctionsFunc

    def accumulateEventTypeStats(self):
        for event_type, count in self.stats_collector.getAllCounters().items():
            if count == 0:
                continue
            self.mp_stats_collector.incrementCounter(event_type, count)
            self.stats_collector.resetCounter(event_type)

    def accumulateReceiveRateStats(self):
        if self.stats_collector.getCounter('events_received') == 0:
            return
        self.mp_stats_collector.incrementCounter('events_received', self.stats_collector.getCounter('events_received'))
        self.stats_collector.resetCounter('events_received')

    def printIntervalStatistics(self):
        self.logger.info("############# Statistics (PID: %s) #############" % os.getpid())
        if self.getConfigurationValue('receive_rate_statistics'):
            self.receiveRateStatistics()
        if self.getConfigurationValue('event_type_statistics'):
            self.eventTypeStatistics()
        #if self.getConfigurationValue('waiting_event_statistics'):
        #    self.eventsInQueuesStatistics()

    def receiveRateStatistics(self):
        self.logger.info(">> Receive rate stats")
        events_received = self.mp_stats_collector.getCounter('events_received')
        self.logger.info("Received events in %ss: %s%s (%s/eps)%s" % (self.getConfigurationValue('interval'), AnsiColors.YELLOW, events_received, (events_received/self.interval), AnsiColors.ENDC))
        if self.emit_as_event:
            self.sendEvent(DictUtils.getDefaultEventDict({"total_count": events_received, "count_per_sec": (events_received/self.interval), "field_name": "all_events", "interval": self.interval}, caller_class_name="Statistics", event_type="statistic"))
        self.mp_stats_collector.resetCounter('events_received')

    def eventTypeStatistics(self):
        self.logger.info(">> EventTypes Statistics")
        for event_type  in sorted(self.mp_stats_collector.getAllCounters().keys()):
            count = self.mp_stats_collector.getCounter(event_type)
            if not event_type.startswith('event_type_'):
                continue
            event_name = event_type.replace('event_type_', '')
            self.logger.info("EventType: %s%s%s - Hits: %s%s%s" % (AnsiColors.YELLOW, event_name, AnsiColors.ENDC, AnsiColors.YELLOW, count, AnsiColors.ENDC))
            if self.emit_as_event:
                self.sendEvent(DictUtils.getDefaultEventDict({"total_count": count, "count_per_sec": (count/self.interval), "field_name": event_name, "interval": self.interval}, caller_class_name="Statistics", event_type="statistic"))
            self.mp_stats_collector.resetCounter(event_type)

    def eventsInQueuesStatistics(self):
        if len(self.module_queues) == 0:
            return
        self.logger.info(">> Queue stats")
        for module_name, queue in sorted(self.module_queues.items()):
            self.logger.info("Events in %s queue: %s%s%s" % (module_name, AnsiColors.YELLOW, queue.qsize(), AnsiColors.ENDC))
            if self.emit_as_event:
                self.sendEvent(DictUtils.getDefaultEventDict({"queue_count": queue.qsize(), "field_name": "queue_counts", "interval": self.interval}, caller_class_name="Statistics", event_type="statistic"))

    def initAfterFork(self):
        # Get all configured queues for waiting event stats.
        self.module_queues = self.lumbermill.getAllQueues()
        TimedFunctionManager.startTimedFunction(self.getRunTimedFunctionsFunc())
        BaseThreadedModule.initAfterFork(self)

    def handleEvent(self, event):
        self.stats_collector.incrementCounter('events_received')
        if self.event_type_statistics:
            try:
                self.stats_collector.incrementCounter('event_type_%s' % event['lumbermill']['event_type'])
            except:
                pass
        yield event

    def shutDown(self):
        self.mp_stats_collector.shutDown()
        BaseThreadedModule.shutDown(self)


# -*- coding: utf-8 -*-
from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class DevNullSink(BaseThreadedModule):
    """
    Just discard messeages send to this module.

    Configuration template:

    - DevNullSink
    """

    module_type = "output"
    """Set module type"""

    def handleEvent(self, event):
        yield None


# -*- coding: utf-8 -*-
import logging
import socket
import sys
import time

from tornado import autoreload
from tornado.iostream import StreamClosedError
from tornado.netutil import bind_sockets
from tornado.tcpserver import TCPServer

import lumbermill.utils.DictUtils as DictUtils
from lumbermill.BaseModule import BaseModule
from lumbermill.utils.Decorators import ModuleDocstringParser


class TornadoTcpServer(TCPServer):

    def __init__(self, io_loop=None, ssl_options=None, gp_module=False, **kwargs):
        self.gp_module = gp_module
        TCPServer.__init__(self, io_loop=io_loop, ssl_options=ssl_options, **kwargs)

    def handle_stream(self, stream, address):
        ConnectionHandler(stream, address, self.gp_module)

class ConnectionHandler(object):

    def __init__(self, stream, address, gp_module):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.gp_module = gp_module
        self.simple_separator = self.gp_module.getConfigurationValue('simple_separator')
        self.regex_separator = self.gp_module.getConfigurationValue('regex_separator')
        self.chunksize = self.gp_module.getConfigurationValue('chunksize')
        self.mode = self.gp_module.getConfigurationValue('mode')
        self.is_open = True
        self.stream = stream
        self.address = address
        (self.host, self.port) = self.address
        self.stream.set_close_callback(self._on_close)
        try:
            if not self.stream.closed():
                if self.mode == 'line' and self.regex_separator:
                    self.stream.read_until_regex(self.regex_separator, self._on_read_line)
                elif self.mode == 'line':
                    self.stream.read_until(self.simple_separator, self._on_read_line)
                else:
                    self.stream.read_bytes(self.chunksize, self._on_read_chunk)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not read from socket %s. Exception: %s, Error: %s." % (self.address, etype, evalue))

    def _on_read_line(self, data):
        data = data.strip()
        if data == "":
            return
        self.sendEvent(data)
        try:
            if not self.stream.reading():
                if self.regex_separator:
                    self.stream.read_until_regex(self.regex_separator, self._on_read_line)
                else:
                    self.stream.read_until(self.simple_separator, self._on_read_line)
        except StreamClosedError:
            pass
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not read from socket %s. Exception: %s, Error: %s." % (self.address, etype, evalue))

    def _on_read_chunk(self, data):
        data = data.strip()
        if data == "":
            return
        self.sendEvent(data)
        try:
            if not self.stream.reading():
                self.stream.read_bytes(self.chunksize, self._on_read_chunk)
        except StreamClosedError:
            pass
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Failed to read from socket %s. Exception: %s, Error: %s." % (self.address, etype, evalue))

    def _on_close(self):
        # Send remaining buffer if neccessary.
        if self.mode == 'stream' and self.stream._read_buffer_size > 0:
            data = ""
            while True:
                try:
                    data += self.stream._read_buffer.popleft().strip()
                except IndexError:
                    if data != "":
                        self.sendEvent(data)
                    break
        self.stream.close()

    def sendEvent(self, data):
        self.gp_module.sendEvent(DictUtils.getDefaultEventDict({"data": data}, caller_class_name="TcpServer", received_from="%s:%d" % (self.host, self.port)))

@ModuleDocstringParser
class TcpServer(BaseModule):
    r"""
    Reads data from tcp socket and sends it to its outputs.
    Should be the best choice perfomancewise if you are on Linux and are running with multiple workers.

    interface:  Ipaddress to listen on.
    port:       Port to listen on.
    timeout:    Sockettimeout in seconds.
    tls:        Use tls or not.
    key:        Path to tls key file.
    cert:       Path to tls cert file.
    mode:       Receive mode, line or stream.
    simple_separator:  If mode is line, set separator between lines.
    regex_separator:   If mode is line, set separator between lines. Here regex can be used.
    chunksize:  If mode is stream, set chunksize in bytes to read from stream.
    max_buffer_size: Max kilobytes to in receiving buffer.

    Configuration template:

    - TcpServer:
       interface:                       # <default: ''; type: string; is: optional>
       port:                            # <default: 5151; type: integer; is: optional>
       timeout:                         # <default: None; type: None||integer; is: optional>
       tls:                             # <default: False; type: boolean; is: optional>
       key:                             # <default: False; type: boolean||string; is: required if tls is True else optional>
       cert:                            # <default: False; type: boolean||string; is: required if tls is True else optional>
       mode:                            # <default: 'line'; type: string; values: ['line', 'stream']; is: optional>
       simple_separator:                # <default: '\n'; type: string; is: optional>
       regex_separator:                 # <default: None; type: None||string; is: optional>
       chunksize:                       # <default: 16384; type: integer; is: optional>
       max_buffer_size:                 # <default: 10240; type: integer; is: optional>
       receivers:
        - NextModule
    """

    module_type = "input"
    """Set module type"""
    can_run_forked = True

    def configure(self, configuration):
        # Call parent configure method
        BaseModule.configure(self, configuration)
        self.server = False
        self.max_buffer_size = self.getConfigurationValue('max_buffer_size') * 10240 #* 10240
        self.start_ioloop = False
        try:
            self.sockets = bind_sockets(self.getConfigurationValue("port"), self.getConfigurationValue("interface"), backlog=128)
            for server_socket in self.sockets:
                server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not listen on %s:%s. Exception: %s, Error: %s." % (self.getConfigurationValue("interface"),
                                                                                        self.getConfigurationValue("port"), etype, evalue))
            self.lumbermill.shutDown()
            return
        autoreload.add_reload_hook(self.shutDown)

    def getStartMessage(self):
        return "listening on %s:%s" % (self.getConfigurationValue("interface"), self.getConfigurationValue("port"))


    def initAfterFork(self):
        BaseModule.initAfterFork(self)
        ssl_options = None
        if self.getConfigurationValue("tls"):
            ssl_options = { 'certfile': self.getConfigurationValue("cert"),
                            'keyfile': self.getConfigurationValue("key")}
        self.server = TornadoTcpServer(ssl_options=ssl_options, gp_module=self, max_buffer_size=self.max_buffer_size)
        self.server.add_sockets(self.sockets)

    def shutDown(self):
        try:
            self.server.stop()
            self.sockets.close()
            # Give os time to free the socket. Otherwise a reload will fail with 'address already in use'
            time.sleep(.2)
        except AttributeError:
            pass



# -*- coding: utf-8 -*-
import sys
import ast
import logging
import datetime
import re

GP_DYNAMIC_VAL_REGEX = re.compile('[\$|%]\((.*?)\)')
GP_DYNAMIC_VAL_REGEX_WITH_TYPES = re.compile('[\$|%]\((.*?)\)(-?\d*[-\.\*]?\d*[sdf]?)')
PYTHON_DYNAMIC_VAL_REGEX = re.compile('%\((.*?)\)')
DYNAMIC_VALUE_REPLACE_PATTERN = r"%(\1)"


# Conditional imports for python2/3
try:
    import __builtin__ as builtins
except ImportError:
    import builtins

def replaceVarsAndCompileString(code_as_string, replacement):
    """
    Parse a string to python code.

    This function will parse the code and replaces all variable names with the replacment.

    Example:

    code_as_string = "matched = VirtualHostName == 'www.gambolutty.com'", replacement = "event['%s']"

     will be parsed and compiled to:
     matched = event['VirtualHostName'] == "www.gambolutty.com"
    """
    try:
        code_ast = ast.parse(code_as_string)
    except:
        etype, evalue, etb = sys.exc_info()
        logging.getLogger("compileStringToConditionalObject").error("%sCould not parse code %s. Exception: %s, Error: %s." % (code_as_string, etype, evalue))
        return False
    transformer = ReplaceVars(replacement)
    transformer.visit(code_ast)
    try:
        code = compile(code_ast, '<ast>', 'exec')
    except:
        etype, evalue, etb = sys.exc_info()
        logging.getLogger("compileStringToConditionalObject").error("%sCould not compile code %s. Exception: %s, Error: %s." % (code_as_string, etype, evalue))
        return False
    return code


class ReplaceVars(ast.NodeTransformer):
    def __init__(self, replacement):
        ast.NodeTransformer.__init__(self)
        self.replacement = replacement

    def visit_Name(self, node):
        # ignore builtins and some other vars
        ignore_nodes = dir(builtins)
        ignore_nodes.extend(["matched", "dependency", "event"])
        if node.id in ignore_nodes:
            return node
        #pprint.pprint(self.mapping % node.id)
        #pprint.pprint(ast.dump(ast.parse(self.mapping % node.id)))
        new_node = ast.parse(self.replacement % node.id).body[0].value
        return new_node

def parseDynamicValuesInString(value):
    """
    Parse a string and replace the configuration notation for dynamic values with pythons notation.
    E.g:
    filter: $(lumbermill.source_module) == 'TcpServer'
    parsed:
    filter: %(lumbermill.source_module)s == 'TcpServer'
    """
    matches = GP_DYNAMIC_VAL_REGEX_WITH_TYPES.search(value)
    if not matches:
        return value
    # Get custom format if set.
    # Defaults to string.
    if matches.group(2):
        replace_pattern = DYNAMIC_VALUE_REPLACE_PATTERN + matches.group(2)
    else:
        replace_pattern = DYNAMIC_VALUE_REPLACE_PATTERN + "s"
    return GP_DYNAMIC_VAL_REGEX_WITH_TYPES.sub(replace_pattern, value)

def parseDynamicValuesInList(value_list, contains_dynamic_value):
    # Copy list since we might change it during iteration.
    value_list_copy = list(value_list)
    for idx, value in enumerate(value_list_copy):
        if isinstance(value, list):
            parseDynamicValuesInList(value_list[idx], contains_dynamic_value)
        elif isinstance(value, dict):
            parseDynamicValuesInDict(value_list[idx], contains_dynamic_value)
        elif isinstance(value, basestring):
            new_value = parseDynamicValuesInString(value)
            if new_value == value:
                continue
            contains_dynamic_value.append(True)
            value_list[idx] = new_value

def parseDynamicValuesInDict(value_dict, contains_dynamic_value):
    # Copy dict since we might change it during iteration.
    value_dict_copy = value_dict.copy()
    for key, value in value_dict_copy.items():
        new_key = key
        if(isinstance(key, basestring)):
            new_key = parseDynamicValuesInString(key)
        if isinstance(value, list):
            parseDynamicValuesInList(value_dict[key], contains_dynamic_value)
        elif isinstance(value, dict):
            parseDynamicValuesInDict(value_dict[key], contains_dynamic_value)
        elif isinstance(value, basestring):
            new_value = parseDynamicValuesInString(value)
            if key == new_key and value == new_value:
                continue
            if new_key != key:
                del value_dict[key]
            contains_dynamic_value.append(True)
            value_dict[new_key] = new_value

def parseDynamicValue(value):
    contains_dynamic_value = []
    if isinstance(value, list):
        parseDynamicValuesInList(value, contains_dynamic_value)
    elif isinstance(value, dict):
        parseDynamicValuesInDict(value, contains_dynamic_value)
    elif isinstance(value, basestring):
        new_value = parseDynamicValuesInString(value)
        if value != new_value:
            value = new_value
            contains_dynamic_value.append(True)
    return {'value': value, 'contains_dynamic_value': bool(contains_dynamic_value)}

def mapDynamicValueInString(value, mapping_dict, use_strftime=False):
    try:
        if use_strftime:
            if sys.platform .startswith('linux'):
                value = datetime.datetime.utcnow().strftime(value)
            # On linux strftime will not remove non-supported format characters.
            # On other platforms this might be different.
            # So we need to take care that other dynamic value notations, which also use the percent sign syntax,
            # are renamed prior to using strftime and renamed to original after strftime was applied.
            # This adds a considerable overhead in parsing.
            # @see: http://bugs.python.org/issue9811
            else:
                value = PYTHON_DYNAMIC_VAL_REGEX.sub(r"$(\1)", value)
                value = datetime.datetime.utcnow().strftime(value)
                value = GP_DYNAMIC_VAL_REGEX.sub(r"%(\1)", value)
        return value % mapping_dict # dot_dict_formatter.format(value, **mapping_dict)
    except KeyError:
        return value
    except (ValueError, TypeError):
        etype, evalue, etb = sys.exc_info()
        logging.getLogger("mapDynamicValueInString").error("Mapping failed for %s. Mapping data: %s. Exception: %s, Error: %s." % (value, mapping_dict, etype, evalue))
        raise

def mapDynamicValueInList(value_list, mapping_dict, use_strftime=False):
    for idx, value in enumerate(value_list):
        if isinstance(value, list):
            mapDynamicValueInList(value, mapping_dict, use_strftime)
        elif isinstance(value, dict):
            mapDynamicValueInDict(value, mapping_dict, use_strftime)
        elif isinstance(value, basestring):
            new_value = mapDynamicValueInString(value, mapping_dict, use_strftime)
            if new_value == value:
                continue
            value_list[idx] = new_value

def mapDynamicValueInDict(value_dict, mapping_dict, use_strftime=False):
    for key, value in value_dict.items():
        new_key = key
        if(isinstance(key, basestring)):
            new_key = mapDynamicValueInString(key, mapping_dict, use_strftime)
        if isinstance(value, list):
            mapDynamicValueInList(value, mapping_dict, use_strftime)
        elif isinstance(value, dict):
            mapDynamicValueInDict(value, mapping_dict, use_strftime)
        elif isinstance(value, basestring):
            new_value = mapDynamicValueInString(value, mapping_dict, use_strftime)
            if key == new_key and value == new_value:
                continue
            value_dict[new_key] = new_value

def mapDynamicValue(value, mapping_dict={}, use_strftime=False):
    if isinstance(value, basestring):
        return mapDynamicValueInString(value, mapping_dict, use_strftime)
    if isinstance(value, list):
        mapDynamicValueInList(value, mapping_dict, use_strftime)
    elif isinstance(value, dict):
        mapDynamicValueInDict(value, mapping_dict, use_strftime)
    return value



# -*- coding: utf-8 -*-
import logging
import logging.handlers
import os
import socket

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser
from lumbermill.utils.DynamicValues import mapDynamicValue


@ModuleDocstringParser
class SyslogSink(BaseThreadedModule):
    """
    Send events to syslog.

    format: Which event fields to send on, e.g. '$(@timestamp) - $(url) - $(country_code)'. If not set the whole event dict is send.
    address: Either a server:port pattern or a filepath to a unix socket, e.g. /dev/log.
    proto: Protocol to use.
    facility: Syslog facility to use. List of possible values, @see: http://epydoc.sourceforge.net/stdlib/logging.handlers.SysLogHandler-class.html#facility_names

    Configuration template:

    - SyslogSink:
       format:                          # <type: string; is: required>
       address:                         # <default: 'localhost:514'; type: string; is: required>
       proto:                           # <default: 'tcp'; type: string; values: ['tcp', 'udp']; is: optional>
       facility:                        # <default: 'user'; type: string; is: optional>
    """

    module_type = "output"
    """Set module type"""

    def configure(self, configuration):
         # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.syslogger = logging.getLogger(self.__class__.__name__)
        self.syslogger.propagate = False
        self.format = self.getConfigurationValue('format')
        if os.path.exists(self.getConfigurationValue('address')):
            address = self.getConfigurationValue('address')
        else:
            server, port = self.getConfigurationValue('address').split(':')
            address = (server, int(port))
        if self.getConfigurationValue('proto') == 'tcp':
            socket_type = socket.SOCK_STREAM
        else:
            socket_type = socket.SOCK_DGRAM
        try:
            facility = logging.handlers.SysLogHandler.facility_names[self.getConfigurationValue('facility')]
        except KeyError:
            self.logger.error("The configured facility %s is unknown." % (self.getConfigurationValue('facility')))
        self.syslog_handler = logging.handlers.SysLogHandler(address, facility=facility, socktype=socket_type)
        self.syslogger.addHandler(self.syslog_handler)
        self.format = self.getConfigurationValue('format')

    def handleEvent(self, event):
        if self.format:
            self.syslogger.info(mapDynamicValue(self.format, event))
        else:
            self.syslogger.info(event)
        yield None

    def shutDown(self):
        self.syslog_handler.close()
        BaseThreadedModule.shutDown(self)



# -*- coding: utf-8 -*-
import Queue
import threading
import types

from dns import resolver, reversename

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Buffers import MemoryCache
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class AddDnsLookup(BaseThreadedModule):
    """
    Add dns info for selected fields. The dns servers used are the ones configured for the system LumberMill is
    running on.

    action: Either resolve or revers.
    source_field: Source field to use for (reverse) lookups.
    target_field: Target field to store result of lookup. If none is provided, the source field will be replaced.
    nameservers: List of nameservers to use. If not provided, the system default servers will be used.
    timeout: Timeout for lookups in seconds.

    Configuration template:

    - AddDnsLookup:
       action:                          # <default: 'resolve'; type: string; is: optional; values: ['resolve', 'reverse']>
       source_field:                    # <default: None; type: string; is: required>
       target_field:                    # <default: None; type: None||string; is: optional>
       nameservers:                     # <default: None; type: None||string||list; is: optional>
       timeout:                         # <default: 1; type: integer; is: optional>
       receivers:
          - NextModule
    """

    module_type = "modifier"
    """Set module type"""

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.in_mem_cache = MemoryCache(size=5000)
        self.lookup_type = self.getConfigurationValue('action')
        self.source_field = self.getConfigurationValue('source_field')
        self.target_field = self.getConfigurationValue('target_field')
        self.nameservers = self.getConfigurationValue('nameservers')
        self.timeout = self.getConfigurationValue('timeout')
        # Allow single string as well.
        if isinstance(self.nameservers, types.StringTypes):
            self.nameservers = [self.nameservers]
        self.lookup_threads_pool_size = 3

    def initAfterFork(self):
        self.resolver = resolver.Resolver()
        self.resolver.timeout = self.timeout
        self.resolver.lifetime = self.timeout
        if self.nameservers:
            self.resolver.nameservers = self.nameservers
        self.queue = Queue.Queue(20)
        self.lookup_threads = [LookupThread(self.queue, self.lookup_type, self) for _ in range(0, self.lookup_threads_pool_size)]
        for thread in self.lookup_threads:
            thread.start()
        BaseThreadedModule.initAfterFork(self)

    def handleEvent(self, event):
        if self.source_field in event:
            target_field = self.target_field if self.target_field else self.source_field
            self.queue.put({'source_field': self.source_field,
                            'target_field': target_field,
                            'event': event})
        yield None

    def shutDown(self):
        BaseThreadedModule.shutDown(self)
        for thread in self.lookup_threads:
            thread.stop()
            thread.join()

class LookupThread(threading.Thread):
    def __init__(self, queue, lookup_type, caller):
        threading.Thread.__init__(self)
        self.queue = queue
        self.lookup_type = lookup_type
        self.caller = caller
        self.daemon = True
        self.alive = True

    def stop(self):
        self.alive = False

    def run(self):
        while self.alive:
            try:
                payload = self.queue.get()
            except Queue.Empty:
                continue
            source_field = payload['source_field']
            target_field = payload['target_field']
            event = payload['event']
            host_or_ip = event[source_field]
            # Try to get it from cache.
            try:
                result = self.caller.in_mem_cache.get(host_or_ip)
            except KeyError:
                if self.lookup_type == 'resolve':
                    result = self.doLookup(host_or_ip)
                elif self.lookup_type == 'reverse':
                    result = self.doReverseLookup(host_or_ip)
            self.caller.in_mem_cache.set(host_or_ip, result)
            event[target_field] = result
            self.caller.sendEvent(event)

    def doReverseLookup(self, ip_address):
        #started = time.time()
        try:
            hostname = str(self.caller.resolver.query(reversename.from_address(ip_address), "PTR")[0])
        except:
            hostname = None
        #if (time.time() - started) > 1:
        #    print("Reverse lookup of %s(%s) took %s." % (ip_address, hostname, time.time() - started))
        return hostname

    def doLookup(self, hostname):
        try:
            ip_address = str(self.caller.resolver.query(hostname)[0])
        except:
            ip_address = None
        return ip_address


# -*- coding: utf-8 -*-
import os
import re
import struct
import sys
from socket import inet_ntoa

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser


@ModuleDocstringParser
class NetFlowParser(BaseThreadedModule):
    r"""
    Netflow parser

    Decode netflow packets.

    source_field:   Input field to decode.
    target_field:   Event field to be filled with the new data.

    Configuration template:

    - NetFlowParser:
       source_field:                    # <default: 'data'; type: string; is: optional>
       target_field:                    # <default: 'data'; type: string; is: optional>
       keep_original:                   # <default: False; type: boolean; is: optional>
       receivers:
        - NextModule
    """

    module_type = "parser"
    """Set module type"""

    NF_V5_HEADER_LENGTH = 24
    NF_V5_RECORD_LENGTH = 48

    TH_FIN = 0x01  # end of data
    TH_SYN = 0x02  # synchronize sequence numbers
    TH_RST = 0x04  # reset connection
    TH_PUSH = 0x08  # push
    TH_ACK = 0x10  # acknowledgment number set
    TH_URG = 0x20  # urgent pointer set
    TH_ECE = 0x40  # ECN echo, RFC 3168
    TH_CWR = 0x80  # congestion window reduced

    IP_PROTOCOLS = {}

    # Helper functions
    def readProtocolInfo(self):
        path = "%s/../assets/ip_protocols" % os.path.dirname(os.path.realpath(__file__))
        r = re.compile("(?P<proto>\S+)\s+(?P<num>\d+)")
        with open(path, 'r') as f:
            for line in f:
                m = r.match(line)
                if not m:
                    continue
                NetFlowParser.IP_PROTOCOLS[int(m.group("num"))] = m.group("proto")

    def configure(self, configuration):
        # Call parent configure method
        BaseThreadedModule.configure(self, configuration)
        self.source_field = self.getConfigurationValue('source_field')
        self.target_field = self.getConfigurationValue('target_field')
        self.drop_original = not self.getConfigurationValue('keep_original')
        self.readProtocolInfo()

    def getTcpFflags(self, flags):
        ret = []
        if flags & NetFlowParser.TH_FIN:
                ret.append('FIN')
        if flags & NetFlowParser.TH_SYN:
                ret.append('SYN')
        if flags & NetFlowParser.TH_RST:
                ret.append('RST')
        if flags & NetFlowParser.TH_PUSH:
                ret.append('PUSH')
        if flags & NetFlowParser.TH_ACK:
                ret.append('ACk')
        if flags & NetFlowParser.TH_URG:
                ret.append('URG')
        if flags & NetFlowParser.TH_ECE:
                ret.append('ECE')
        if flags & NetFlowParser.TH_CWR:
                ret.append('CWR')
        return ret

    def decodeVersion5(self, raw_nf_data, record_count):
        nf_data = {}
        (nf_data['sys_uptime'], nf_data['unix_secs'], nf_data['unix_nsecs'], nf_data['flow_sequence'], nf_data['engine_type'], nf_data['engine_id'], nf_data['sampling_interval']) = struct.unpack('!IIIIBBH', raw_nf_data[4:24])
        for i in xrange(0, record_count):
            record_starts_at = NetFlowParser.NF_V5_HEADER_LENGTH + (i * NetFlowParser.NF_V5_RECORD_LENGTH)
            record = raw_nf_data[record_starts_at:record_starts_at+NetFlowParser.NF_V5_RECORD_LENGTH]
            # Decode record, except src and dest addresses.
            decoded_record = struct.unpack('!HHIIIIHHBBBBHHBBH', record[12:])
            nf_data['srcaddr'] = inet_ntoa(record[:4])
            nf_data['dstaddr'] = inet_ntoa(record[4:8])
            nf_data['nexthop'] = inet_ntoa(record[8:12])
            nf_data['snmp_index_in_interface'] = decoded_record[0]
            nf_data['snmp_index_out_interface'] = decoded_record[1]
            nf_data['packet_count'] = decoded_record[2]
            nf_data['byte_count'] = decoded_record[3]
            nf_data['uptime_start'] = decoded_record[4]
            nf_data['uptime_end'] = decoded_record[5]
            nf_data['srcport'] = decoded_record[6]
            nf_data['dstport'] = decoded_record[7]
            nf_data['tcp_flags_binary'] = decoded_record[9]
            nf_data['tcp_flags'] = self.getTcpFflags(decoded_record[9])
            nf_data['prot'] = decoded_record[10]
            nf_data['prot_name'] = NetFlowParser.IP_PROTOCOLS[decoded_record[10]]
            nf_data['tos'] = decoded_record[11]
            nf_data['src_as'] = decoded_record[12]
            nf_data['dst_as'] = decoded_record[13]
            nf_data['src_mask'] = decoded_record[14]
            nf_data['dst_mask'] = decoded_record[15]
            yield nf_data

    def handleEvent(self, event):
        if self.source_field not in event:
            yield event
            return
        raw_nf_data = event[self.source_field]
        # Try to get netflow version.
        try:
             (version, record_count) = struct.unpack('!HH', raw_nf_data[0:4])
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.warning("Could not detect netflow version: %s. Exception: %s, Error: %s." % (raw_nf_data, etype, evalue))
            yield event
        # Call decoder for detected version.
        try:
            decoder_func = getattr(self, "decodeVersion%s" % version)
        except AttributeError:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Netflow parser does not implement decoder for netflow version: %s. Exception: %s, Error: %s" % (version, etype, evalue))
            self.lumbermill.shutDown()
        copy_event = False
        for netflow_data in decoder_func(raw_nf_data, record_count):
            if copy_event:
                event = event.copy()
            copy_event = True
            if self.drop_original and self.source_field is not self.target_field:
                event.pop(self.source_field, None)
            event.update({self.target_field: netflow_data})
            event['lumbermill']['event_type'] = "NetFlowV%s" % version
            yield event





#!/usr/bin/env python
#
# Copyright 2009 Phus Lu
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Blocking and non-blocking Redis client implementations using IOStream."""

import logging
import socket
from collections import deque
from tornado.ioloop import IOLoop
from tornado.iostream import IOStream


def encode(request):
    """Encode request (command, *args) to redis bulk bytes.

    Note that command is a string defined by redis.
    All elements in args should be a string.
    """
    assert isinstance(request, tuple)
    data = '*%d\r\n' % len(request) + ''.join(['$%d\r\n%s\r\n' % (len(str(x)), x) for x in request])
    return data

def decode(read_buffer):
    """Decode redis buffer to python object."""
    assert isinstance(read_buffer, deque)
    first_line = read_buffer.popleft()
    c = first_line[0]
    if c == '+':
        return first_line[1:-2]
    elif c == '-':
        return RedisError(first_line[1:-2], data)
    elif c == ':':
        return int(first_line[1:])
    elif c == '$':
        number = int(first_line[1:])
        if number == -1:
            return None
        else:
            return read_buffer[0][:number]
    elif c == '*':
        number = int(first_line[1:])
        if number == -1:
            return None
        else:
            result = []
            while number:
                line = read_buffer.popleft()
                c = line[0]
                if c == '$':
                    length  = int(line[1:])
                    element = read_buffer.popleft()[:length]
                    result.append(element)
                else:
                    if c == ':':
                        element = int(line[1:])
                    else:
                        element = read_buffer.popleft()[:-2]
                    result.append(element)
                number -= 1
            return result
    else:
        raise RedisError('bulk cannot startswith %r' % c, data)

class RedisClient(object):
    """A blocking Redis client.

    This interface is provided for convenience and testing; most applications
    that are running an IOLoop will want to use `AsyncRedisClient` instead.
    Typical usage looks like this::

        redis_client = redisclient.RedisClient(('127.0.0.1', 6379))
        try:
            result = redis_client.fetch(('set', 'foo', 'bar'))
            print result
        except redisclient.RedisError, e:
            print "Error:", e
    """
    def __init__(self, address):
        self.address = address
        self._io_loop = IOLoop()
        self._async_client = AsyncRedisClient(self.address, self._io_loop)
        self._result = None
        self._closed = False

    def __del__(self):
        self.close()

    def close(self):
        """Closes the RedisClient, freeing any resources used."""
        if not self._closed:
            self._async_client.close()
            self._io_loop.close()
            self._closed = True

    def fetch(self, request, **kwargs):
        """Executes a request, returning an `result`.

        The request may be a tuple object. like ('set','foo','bar')

        If an error occurs during the fetch, we raise an `RedisError`.
        """
        def callback(result):
            self._result = result
            self._io_loop.stop()
        self._async_client.fetch(request, callback, **kwargs)
        self._io_loop.start()
        result = self._result
        self._result = None
        return result

class AsyncRedisClient(object):
    """An non-blocking Redis client.

    Example usage::

        import ioloop

        def handle_request(result):
            print 'Redis reply: %r' % result
            ioloop.IOLoop.instance().stop()

        redis_client = AsyncRedisClient(('127.0.0.1', 6379))
        redis_client.fetch(('set', 'foo', 'bar'), None)
        redis_client.fetch(('get', 'foo'), handle_request)
        ioloop.IOLoop.instance().start()

    This class implements a Redis client on top of Tornado's IOStreams.
    It does not currently implement all applicable parts of the Redis
    specification, but it does enough to work with major redis server APIs
    (mostly tested against the LIST/HASH/PUBSUB API so far).

    This class has not been tested extensively in production and
    should be considered somewhat experimental as of the release of
    tornado 1.2.  It is intended to become the default tornado
    AsyncRedisClient implementation.
    """

    def __init__(self, address, io_loop=None, socket_timeout=10):
        """Creates a AsyncRedisClient.

        address is the tuple of redis server address that can be connect by
        IOStream. It can be to ('127.0.0.1', 6379).
        """
        self.address         = address
        self.io_loop         = io_loop or IOLoop.instance()
        self._callback_queue = deque()
        self._callback       = None
        self._read_buffer    = None
        self._result_queue   = deque()
        self.socket          = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(socket_timeout)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.stream          = IOStream(self.socket, self.io_loop)
        self.stream.connect(self.address, self._wait_result)

    def close(self):
        """Destroys this redis client, freeing any file descriptors used.
        Not needed in normal use, but may be helpful in unittests that
        create and destroy redis clients.  No other methods may be called
        on the AsyncRedisClient after close().
        """
        self.stream.close()

    def fetch(self, request, callback):
        """Executes a request, calling callback with an redis `result`.

        The request should be a string tuple. like ('set', 'foo', 'bar')

        If an error occurs during the fetch, a `RedisError` exception will
        throw out. You can use try...except to catch the exception (if any)
        in the callback.
        """
        self._callback_queue.append(callback)
        self.stream.write(encode(request))

    def _wait_result(self):
        """Read a completed result data from the redis server."""
        self._read_buffer = deque()
        self.stream.read_until('\r\n', self._on_read_first_line)

    def _maybe_callback(self):
        """Try call callback in _callback_queue when we read a redis result."""
        try:
            read_buffer    = self._read_buffer
            callback       = self._callback
            result_queue   = self._result_queue
            callback_queue = self._callback_queue
            if result_queue:
                result_queue.append(read_buffer)
                read_buffer = result_queue.popleft()
            if callback_queue:
                callback = self._callback = callback_queue.popleft()
            if callback:
                callback(decode(read_buffer))
        except Exception:
            logging.error('Uncaught callback exception', exc_info=True)
            self.close()
            raise
        finally:
            self._wait_result()

    def _on_read_first_line(self, data):
        self._read_buffer.append(data)
        c = data[0]
        if c in ':+-':
            self._maybe_callback()
        elif c == '$':
            if data[:3] == '$-1':
                self._maybe_callback()
            else:
                length = int(data[1:])
                self.stream.read_bytes(length+2, self._on_read_bulk_body)
        elif c == '*':
            if data[1] in '-0' :
                self._maybe_callback()
            else:
                self._multibulk_number = int(data[1:])
                self.stream.read_until('\r\n', self._on_read_multibulk_bulk_head)

    def _on_read_bulk_body(self, data):
        self._read_buffer.append(data)
        self._maybe_callback()

    def _on_read_multibulk_bulk_head(self, data):
        self._read_buffer.append(data)
        c = data[0]
        if c == '$':
            length = int(data[1:])
            self.stream.read_bytes(length+2, self._on_read_multibulk_bulk_body)
        else:
            self._maybe_callback()

    def _on_read_multibulk_bulk_body(self, data):
        self._read_buffer.append(data)
        self._multibulk_number -= 1
        if self._multibulk_number:
            self.stream.read_until('\r\n', self._on_read_multibulk_bulk_head)
        else:
            self._maybe_callback()

class RedisError(Exception):
    """Exception thrown for an unsuccessful Redis request.

    Attributes:

    data - Redis error data error code, e.g. -(ERR).
    """
    def __init__(self, message, data=None):
        self.data = data
        Exception.__init__(self, '(Error): %s' % message)


def test():
    def handle_request(result):
        print 'Redis reply: %r' % result
    redis_client = AsyncRedisClient(('localhost', 6379))
    redis_client.fetch(('subscribe', 'GamboPutty'), handle_request)
    #redis_client.fetch(('set', 'foo', 'bar'), handle_request)
    #redis_client.fetch(('get', 'foo'), lambda x:(handle_request(x),IOLoop.instance().stop()))
    IOLoop.instance().start()

if __name__ == '__main__':
    test()


# -*- coding: utf-8 -*-
import collections
import hashlib
import os
import re
import socket
import sys
import threading
import time

import msgpack
from Crypto import Random
from Crypto.Cipher import AES

from lumbermill.BaseThreadedModule import BaseThreadedModule
from lumbermill.utils.Decorators import ModuleDocstringParser, setInterval
from lumbermill.utils.misc import TimedFunctionManager


class PackMember:

    def __init__(self, host, hostname=None):
        self.host = host
        self.ip_address = host[0]
        self.port = host[1]
        self.hostname = hostname
        self.last_seen = time.time()

    def getHost(self):
        return self.host

    def getIp(self):
        return self.ip_address

    def getPort(self):
        return self.port

    def getHostName(self):
        if not self.hostname:
            try:
                self.hostname = socket.gethostbyaddr(self.ip_address)
            except:
                self.hostname = self.ip_address
        return self.hostname

    def updateLastSeen(self):
        self.last_seen = time.time()

    def isAlive(self):
        return True if time.time() - self.last_seen < 30 else False



@ModuleDocstringParser
class Pack(BaseThreadedModule):
    """
    Pack base module. Handles pack leader discovery and alive checks of pack followers.

    IMPORTANT:
    This is just a first alpha implementation. No leader election, no failover, no sanity checks for conflicting leaders.

    name: Name of the cluster. Used for auto-discovery in same network.
    secret: pre shared key to en/decrypt cluster messages.
    broadcast: Ipaddress for udp broadcasts.
    interface:  Ipaddress to listen on.
    port:   Port to listen on.
    interval: Autodiscover interval.
    pack: Set this node to be either leader or member.

    Configuration template:

    - Pack:
       name:                            # <type: string; is: required>
       secret:                          # <type: string; is: required>
       broadcast:                       # <type: string; is: required>
       interface:                       # <default: '0.0.0.0'; type: string; is: optional>
       port:                            # <default: 5252; type: integer; is: optional>
       interval:                        # <default: 10; type: integer; is: optional>
       pack:                            # <default: 'leader'; type: string; values: ['leader', 'follower']; is: optional>
    """

    module_type = "stand_alone"
    """ Set module type. """

    can_run_forked = False

    def configure(self, configuration):
        # self.logger.setLevel(logging.DEBUG)
        # Call parent configure method.
        BaseThreadedModule.configure(self, configuration)
        self.is_leader = True if self.getConfigurationValue('pack') == 'leader' else False
        self.pack_followers = {}
        self.cluster_name = self.getConfigurationValue('name')
        self.discovered_leader = None
        self.secret = hashlib.sha256(self.getConfigurationValue('secret')).digest()
        self.handlers = collections.defaultdict(list)
        self.lock = threading.Lock()
        # Setup socket.
        self.interface_addr = self.getConfigurationValue('interface')
        self.interface_port = self.getConfigurationValue('port')
        self.broadcast_addr = self.getConfigurationValue('broadcast')
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self.socket.settimeout(1)
        self.hostname = socket.gethostname()
        try:
            self.socket.bind((self.interface_addr, self.interface_port))
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.error("Could not listen on %s:%s. Exception: %s, Error: %s." % (self.getConfigurationValue("interface"),
                                                                                        self.getConfigurationValue("port"), etype, evalue))
            self.alive = False
            self.lumbermill.shutDown()
            return
        self.addHandlers()

    def addHandlers(self):
        handle_marker = 'leaderHandle' if self.is_leader else 'followerHandle'
        for method_name in dir(self):
            if not method_name.startswith(handle_marker):
                continue
            action_name = method_name.replace(handle_marker, '')
            action_name = self.convertCamelCaseToUnderScore(action_name)
            self.addHandler(action_name, getattr(self, method_name))

    def convertCamelCaseToUnderScore(self, camel_case):
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', camel_case)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

    def encrypt(self, message):
        x = AES.block_size - len(message) % AES.block_size
        message = message + (chr(x) * x)
        iv = Random.OSRNG.posix.new().read(AES.block_size)
        cipher = AES.new(self.secret, AES.MODE_CBC, iv)
        return iv + cipher.encrypt(message)

    def decrypt(self, message):
        unpad = lambda s: s[:-ord(s[-1])]
        iv = message[:AES.block_size]
        cipher = AES.new(self.secret, AES.MODE_CBC, iv)
        return unpad(cipher.decrypt(message))[AES.block_size:]

    def getDefaultMessageDict(self, action, custom_dict={}):
        message_dict = {'action': action, 'sender': self.hostname, 'cluster': self.cluster_name, 'timestamp': time.time()}
        message_dict.update(custom_dict)
        return message_dict

    def getPackMembers(self):
        return self.pack_followers

    def getDiscoveredLeader(self):
        return self.discovered_leader

    def sendBroadcastMessage(self, message):
         # msgpack encode end encrypt message.
        str_msg = msgpack.packb(message)
        str_msg = self.encrypt(str_msg)
        try:
            self.socket.sendto(str_msg, (self.broadcast_addr, self.interface_port))
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.warning("Could not send broadcast message %s to %s. Exception: %s, Error: %s." % (message, host, etype, evalue))

    def sendMessage(self, message, ip_address):
        self.logger.debug("Sending message %s to %s." % (message, ip_address))
         # Json encode end encrypt message.
        str_msg = msgpack.packb(message)
        str_msg = self.encrypt(str_msg)
        try:
            self.socket.sendto(str_msg, (ip_address, self.interface_port))
        except:
            etype, evalue, etb = sys.exc_info()
            self.logger.warning("Could not send message %s to %s. Exception: %s, Error: %s." % (message, ip_address, etype, evalue))

    def sendMessageToPackFollower(self, pack_member, message):
        if pack_member not in self.pack_followers.values():
            self.logger.warning('Can not send message to unknown pack follower %s.' % pack_member)
            return
        self.sendMessage(message, pack_member.getIp())

    def sendMessageToPack(self, message):
        for pack_member in self.pack_followers.itervalues():
            self.sendMessageToPackFollower(pack_member, message)

    def sendMessageToPackLeader(self, message):
        self.sendMessage(message, self.discovered_leader.getIp())


    def addHandler(self, action, callback):
        self.logger.debug('Adding handler %s for %s.' % (callback, action))
        self.handlers[action].append(callback)

    def run(self):
        if self.is_leader:
            TimedFunctionManager.startTimedFunction(self.sendAliveRequests)
            TimedFunctionManager.startTimedFunction(self.sendDiscoverBroadcast)
            TimedFunctionManager.startTimedFunction(self.dropDeadPackFollowers)
        else:
            TimedFunctionManager.startTimedFunction(self.dropDeadPackLeader)
        self.logger.info("Running as pack %s." % self.getConfigurationValue('pack'))
        while self.alive:
            try:
                message, host = self.socket.recvfrom(64536)
            except socket.timeout:
                continue
            try:
                # Decrypt and msgpack decode message.
                message = msgpack.unpackb(self.decrypt(message))
            except:
                etype, evalue, etb = sys.exc_info()
                self.logger.warning("Could not parse cluster message %s. Maybe your secrets differ? Exception: %s, Error: %s." % (message, etype, evalue))
                continue
            # Ignore messages for other clusters and from self.
            if message['cluster'] != self.cluster_name or message['sender'] == self.hostname:
                continue
            self.logger.info("Received message %s from %s." % (message, host[0]))
            if message['action'] not in self.handlers.keys():
                self.logger.warning('Got request for unknown handler %s.' % message['action'])
            # Excecute callbacks
            for callback in self.handlers["%s" % message['action']]:
                self.logger.debug('Calling callback %s for %s.' % (callback, message['action']))
                callback(host, message)

    @setInterval(10, call_on_init=True)
    def sendDiscoverBroadcast(self):
        """
        Leader sends udp broadcast messages every 10 seconds.
        """
        message = self.getDefaultMessageDict(action='discovery_request')
        self.logger.debug('Sending broadcast pack follower discovery.')
        self.sendBroadcastMessage(message)

    @setInterval(10)
    def sendAliveRequests(self):
        """
        Leader sends alive requests to all dicovered followers every 10 seconds.
        """
        for ip_address, pack_follower in self.pack_followers.items():
            message = self.getDefaultMessageDict(action='alive_request')
            self.sendMessageToPackFollower(pack_follower, message)

    @setInterval(10)
    def dropDeadPackFollowers(self):
        """
        Drop followers who have not been seen the last 30 seconds.
        @see PackMember.isAlive() and PackMember.updateLastSeen()
        """
        with self.lock:
            for ip_address, pack_follower in self.pack_followers.items():
                if pack_follower.isAlive():
                    continue
                self.logger.warning('Dropping dead pack follower %s, %s.' % (pack_follower.getHostName(), ip_address))
                self.pack_followers.pop(ip_address)

    @setInterval(1)
    def dropDeadPackLeader(self):
        """
        Drop leader who has not been seen the last 30 seconds.
        @see PackMember.isAlive() and PackMember.updateLastSeen()
        """
        if not self.discovered_leader:
            return
        with self.lock:
            if self.discovered_leader.isAlive():
                return
            self.logger.warning('Dropping dead pack leader %s, %s.' % (self.discovered_leader.getHostName(), self.discovered_leader.getIp()))
            self.discovered_leader = None

    """
    Pack follower discovery procedure.

    - leader sends a broadcast 'discovery_request' message
    - follower replies with a 'discovery_reply' message
    - leader replies with a 'discovery_finish' message
    - follower sets its discovered leader
    """
    def followerHandleDiscoveryRequest(self, sender, message):
        sender_ip = sender[0]
        if self.discovered_leader and self.discovered_leader.getIp() == sender_ip:
            return
        message = self.getDefaultMessageDict(action='discovery_reply')
        self.sendMessage(message, sender[0])

    def leaderHandleDiscoveryReply(self, sender, message):
        self.logger.info("Discovered %s as pack member in pid %s." % (message['sender'], os.getpid()))
        sender_ip = sender[0]
        try:
            pack_member = self.pack_followers[sender_ip]
        except KeyError:
            pack_member = PackMember(sender, message['sender'])
            self.pack_followers[sender_ip] = pack_member
        message = self.getDefaultMessageDict(action='discovery_finish')
        self.sendMessage(message, pack_member.getIp())

    def followerHandleDiscoveryFinish(self, sender, message):
        self.logger.info("Discovered %s as pack leader." % message['sender'])
        self.discovered_leader = PackMember(sender, message['sender'])

    """
    Pack follower alive procedure.

    - leader sends an 'alive_request' message
    - follower replies with a 'alive_reply' message
    - leader calls follower.updateLastSeen() and replies with a 'alive_finish' message
    - follower calls leader.updateLastSeen()
    """
    def followerHandleAliveRequest(self, sender, message):
        sender_ip = sender[0]
        if not self.discovered_leader or sender_ip != self.discovered_leader.getIp():
            return
        self.logger.debug('Got alive request from %s.' % sender_ip)
        message = self.getDefaultMessageDict(action='alive_reply')
        self.sendMessageToPackLeader(message)

    def leaderHandleAliveReply(self, sender, message):
        try:
            pack_follower = self.pack_followers[sender[0]]
        except KeyError:
            return
        self.logger.debug('Got alive reply from %s.' % pack_follower.getHostName())
        with self.lock:
            pack_follower.updateLastSeen()
        message = self.getDefaultMessageDict(action='alive_finish')
        self.sendMessageToPackFollower(pack_follower, message)

    def followerHandleAliveFinish(self, sender, message):
        sender_ip = sender[0]
        if not self.discovered_leader or sender_ip != self.discovered_leader.getIp():
            return
        self.discovered_leader.updateLastSeen()

    def shutDown(self):
        # Call parent configure method.
        BaseThreadedModule.shutDown(self)
        self.socket.close()
        self.socket = None

